import re
from typing import Any, Optional

from rimu import options


class Expand:
    '''Processing priority(highest to lowest): container, skip, spans and specials.
       If spans is true then both spans and specials are processed.
       They are assumed false if they are not explicitly defined.
       If a custom filter is specified their use depends on the filter.'''
    macros: Optional[bool]
    container: Optional[bool]
    skip: Optional[bool]
    spans: Optional[bool]
    # Span substitution also expands special characters.
    specials: Optional[bool]

    def __init__(self,
                 macros: bool = None,
                 container: bool = None,
                 skip: bool = None,
                 spans: bool = None,
                 specials: bool = None,
                 ):
        self.macros = macros
        self.container = container
        self.skip = skip
        self.spans = spans
        self.specials = specials

    def __eq__(self, obj: Any) -> bool:
        return (isinstance(obj, Expand) and
                self.macros == obj.macros and
                self.container == obj.container and
                self.skip == obj.skip and
                self.spans == obj.spans and
                self.specials == obj.specials)

    @classmethod
    def copyFrom(cls, other: 'Expand') -> 'Expand':
        return Expand(
            other.macros,
            other.container,
            other.skip,
            other.spans,
            other.specials,
        )

    def merge(self, other: 'Expand') -> None:
        if other is None:
            return
        if other.macros is not None:
            self.macros = other.macros
        if other.container is not None:
            self.container = other.container
        if other.skip is not None:
            self.skip = other.skip
        if other.spans is not None:
            self.spans = other.spans
        if other.specials is not None:
            self.specials = other.specials

    def parse(self, opts: str) -> None:
        '''Parse block-options string into blockOptions.'''
        if opts != '':
            for opt in re.split(r'\s+', opts.strip()):
                if options.isSafeModeNz() and opt == '-specials':
                    options.errorCallback('-specials block option not valid in safeMode')
                    continue
                if re.match(r'^[+-](macros|spans|specials|container|skip)$', opt) is not None:
                    value = opt[0] == '+'
                    if opt[1:] == 'macros':
                        self.macros = value
                    elif opt[1:] == 'spans':
                        self.spans = value
                    elif opt[1:] == 'specials':
                        self.specials = value
                    elif opt[1:] == 'container':
                        self.container = value
                    elif opt[1:] == 'skip':
                        self.skip = value
                else:
                    options.errorCallback('illegal block option: ' + opt)



import re
from typing import List, Optional, Pattern


class Def:
    quote: str
    opentTag: str
    closeTag: str
    spans: bool

    def __init__(self,
                 quote: str,
                 openTag: str,
                 closeTag: str,
                 spans: bool,
                 ) -> None:
        self.quote = quote
        self.openTag = openTag
        self.closeTag = closeTag
        self.spans = spans

    @classmethod
    def copyFrom(cls, d: 'Def') -> 'Def':
        return Def(d.quote, d.openTag, d.closeTag, d.spans)


defs: List[Def] = []  # Mutable definitions initialized by DEFAULT_DEFS.

DEFAULT_DEFS: List[Def] = [
    Def(quote='**', openTag='<strong>', closeTag='</strong>', spans=True),
    Def(quote='*', openTag='<em>', closeTag='</em>', spans=True),
    Def(quote='__', openTag='<strong>', closeTag='</strong>', spans=True),
    Def(quote='_', openTag='<em>', closeTag='</em>', spans=True),
    Def(quote='``', openTag='<code>', closeTag='</code>', spans=False),
    Def(quote='`', openTag='<code>', closeTag='</code>', spans=False),
    Def(quote='~~', openTag='<del>', closeTag='</del>', spans=True),
]

quotesRe: Pattern[str]      # Searches for quoted text.
unescapeRe: Pattern[str]    # Searches for escaped quotes.


def init() -> None:
    '''Reset definitions to defaults.'''
    global defs
    # Make shallow copy of DEFAULT_DEFS (list and list objects).
    defs.clear()
    defs.extend(map(lambda d: Def.copyFrom(d), DEFAULT_DEFS))
    initializeRegExps()


def initializeRegExps() -> None:
    '''Synthesise re's to find and unescape quotes.'''
    global quotesRe, unescapeRe
    quotes = list(map(lambda d: re.escape(d.quote), defs))
    # $1 is quote character(s), $2 is quoted text.
    # Quoted text cannot begin or end with whitespace.
    # Quoted can span multiple lines.
    # Quoted text cannot end with a backslash.
    quotesRe = re.compile(
        r'\\?(' + '|'.join(quotes) + r')([^\s\\]|\S[\s\S]*?[^\s\\])\1')
    # $1 is quote character(s).
    unescapeRe = re.compile(r'\\(' + '|'.join(quotes) + ')')


def getDefinition(quote: str) -> Optional[Def]:
    '''Return the quote definition corresponding to 'quote' character, return null if not found.'''
    for d in defs:
        if d.quote == quote:
            return d
    return None


def setDefinition(qdef: Def) -> None:
    '''Update existing or add new quote definition.'''
    d = getDefinition(qdef.quote)
    if d is not None:
        # Update existing definition.
        d.openTag = qdef.openTag
        d.closeTag = qdef.closeTag
        d.spans = qdef.spans
    else:
        # Double-quote definitions are prepended to the array so they are matched
        # before single-quote definitions(which are appended to the array).
        if len(qdef.quote) == 2:
            defs.insert(0, qdef)
        else:
            defs.append(qdef)
        initializeRegExps()


def unescape(s: str) -> str:
    '''Strip backslashes from quote characters.'''
    return unescapeRe.sub(lambda m: m[1], s)



import re
from typing import Match

from rimu import macros, options, spans
from rimu.expansion import Expand


def replaceSpecialChars(s: str) -> str:
    return s.replace('&', '&amp;').replace('>', '&gt;').replace('<', '&lt;')


def replaceMatch(match: Match, replacement: str, expand: Expand = None) -> str:
    '''Replace pattern '$1' or '$$1', '$2' or '$$2'... in `replacement` with corresponding match groups
       from `match`. If pattern starts with one '$' character add specials to `expand`,
       if it starts with two '$' characters add spans to `expand`.'''
    if expand is None:
        expand = Expand()

    def repl(m):
        # Replace $1, $2 ... with corresponding match groups.
        if m[1] == '$$':
            expand.spans = True
        else:
            expand.specials = True
        i = int(m[2])
        # match group number.
        if i > match.re.groups:
            options.errorCallback('undefined replacement group: ' + m[0])
            return ''
        result = match[i]
        # match group text.
        return replaceInline(result, expand)
    return re.sub(r'(\${1,2})(\d)', repl, replacement)


def replaceInline(text: str, expand: Expand) -> str:
    '''Replace the inline elements specified in options in text and return the result.'''
    result = text
    if expand.macros:
        result = macros.render(result)
    # Spans also expand special characters.
    if expand.spans:
        result = spans.render(result)
    elif expand.specials:
        result = replaceSpecialChars(result)
    return result



import re
from typing import Callable, List, Match, Optional, Pattern

from rimu import options, utils

Filter = Optional[Callable[[Match[str], 'Def'], str]]


class Def:
    match: Pattern[str]
    replacement: str
    filter: Filter

    def __init__(self,
                 match: Pattern[str],
                 replacement: str,
                 filter: Filter = None,
                 ):
        self.match = match
        self.replacement = replacement
        self.filter = filter

    @classmethod
    def copyFrom(cls, d: 'Def') -> 'Def':
        return Def(d.match, d.replacement, d.filter)


defs: List[Def] = []  # Mutable definitions initialized by DEFAULT_DEFS.

DEFAULT_DEFS: List[Def] = [
    # Begin match with \\? to allow the replacement to be escaped.
    # Global flag must be set on match re's so that the RegExp lastIndex property is set.
    # Replacements and special characters are expanded in replacement groups($1..).
    # Replacement order is important.

    # DEPRECATED as of 3.4.0.
    # Anchor: << #id>>
    Def(
        match=re.compile(r'\\?<<#([a-zA-Z][\w\-]*)>>'),
        replacement='<span id="$1"></span>',
        filter=lambda match, d:
            '' if options.skipBlockAttributes() else utils.replaceMatch(match, d.replacement)
    ),

    # Image: <image:src|alt>
    # src= $1, alt = $2
    Def(
        match=re.compile(r'\\?<image:([^\s|]+)\|(.*?)>', re.DOTALL),
        replacement='<img src="$1" alt="$2">'),

    # Image: <image:src>
    # src= $1, alt = $1
    Def(
        match=re.compile(r'\\?<image:([^\s|]+?)>'),
        replacement='<img src="$1" alt="$1">'),

    # Image: ![alt](url)
    # alt= $1, url = $2
    Def(
        match=re.compile(r'\\?!\[([^[]*?)]\((\S+?)\)'),
        replacement='<img src="$2" alt="$1">'),

    # Email: <address|caption>
    # address= $1, caption = $2
    Def(
        match=re.compile(r'\\?<(\S+@[\w.\-]+)\|(.+?)>', re.DOTALL),
        replacement='<a href="mailto:$1">$$2</a>'),

    # Email: <address>
    # address= $1, caption = $1
    Def(
        match=re.compile(r'\\?<(\S+@[\w.\-]+)>'),
        replacement='<a href="mailto:$1">$1</a>'),

    # Open link in new window: ^[caption](url)
    # caption= $1, url = $2
    Def(
        match=re.compile(r'\\?\^\[([^[]*?)]\((\S+?)\)'),
        replacement='<a href="$2" target="_blank">$$1</a>'),

    # Link: [caption](url)
    # caption= $1, url = $2
    Def(
        match=re.compile(r'\\?\[([^[]*?)]\((\S+?)\)'),
        replacement='<a href="$2">$$1</a>'),

    # Link: <url|caption>
    # url= $1, caption = $2
    Def(
        match=re.compile(r'\\?<(\S+?)\|(.*?)>', re.DOTALL),
        replacement='<a href="$1">$$2</a>'),

    # HTML inline tags.
    # Match HTML comment or HTML tag.
    # $1= tag, $2 = tag name
    Def(
        match=re.compile(
            r'\\?(<!--(?:[^<>&]*)?-->|<\/?([a-z][a-z0-9]*)(?:\s+[^<>&]+)?>)', re.IGNORECASE),
        replacement='',
        # Matched HTML comment or inline tag.
        filter=lambda match, d: options.htmlSafeModeFilter(match[1])
    ),

    # Link: <url>
    # url= $1
    Def(match=re.compile(r'\\?<([^|\s]+?)>'),
        replacement='<a href="$1">$1</a>'),

    # Auto-encode(most) raw HTTP URLs as links.
    Def(
        match=re.compile(
            r'\\?((?:http|https):\/\/[^\s"' + r"']*[A-Za-z0-9/#])"),
        replacement='<a href="$1">$1</a>'),

    # Character entity.
    Def(
        match=re.compile(r'\\?(&[\w#][\w]+;)'),
        replacement='',
        filter=lambda match, d: match[1]  # Pass the entity through verbatim.
    ),

    # Line-break (space followed by \ at end of line).
    Def(match=re.compile(r'[\\ ]\\(\n|$)'), replacement='<br>$1'),

    # This hack ensures backslashes immediately preceding closing code quotes are rendered
    # verbatim(Markdown behaviour).
    # Works by finding escaped closing code quotes and replacing the backslash and the character
    # preceding the closing quote with itself.
    Def(match=re.compile(r'(\S\\)(?=`)'), replacement='$1'),

    # This hack ensures underscores within words rendered verbatim and are not treated as
    # underscore emphasis quotes(GFM behaviour).
    Def(match=re.compile(r'([a-zA-Z0-9]_)(?=[a-zA-Z0-9])'), replacement='$1'),
]

# Reset definitions to defaults.


def init() -> None:
    '''Reset definitions to defaults.'''
    global defs
    # Make shallow copy of DEFAULT_DEFS (list and list objects).
    defs.clear()
    for d in DEFAULT_DEFS:
        defs.append(Def.copyFrom(d))


def getDefinition(pattern: str) -> Optional[Def]:
    '''Return the replacment definition matching the regular expresssion pattern, return null if not found.'''
    for d in defs:
        if d.match.pattern == pattern:
            return d
    return None


def setDefinition(pattern: str, flags: str, replacement: str) -> None:
    '''Update existing or add new replacement definition.'''
    # Flag properties are read-only so have to create new RegExp.
    flgs = 0
    if 'i' in flags:
        flgs |= re.IGNORECASE
    if 'm' in flags:
        flgs |= re.MULTILINE
    regexp = re.compile(pattern, flgs)
    d = getDefinition(pattern)
    if d is not None:
        # Update existing definition.
        d.match = regexp
        d.replacement = replacement
    else:
        # Append new definition to end of defs list(custom definitons have lower precedence).
        defs.append(Def(match=regexp, replacement=replacement))



import re
from typing import List

from rimu import options, utils
from rimu.expansion import Expand

classes: str  # Space separated HTML class names.
id: str  # HTML element id.
css: str  # HTML CSS styles.
attributes: str  # Other HTML element attributes.
opts: Expand

ids: List[str] = []  # List of allocated HTML ids.


def init() -> None:
    global classes, id, css, attributes, opts, ids
    classes = ''
    id = ''
    css = ''
    attributes = ''
    opts = Expand()
    ids.clear()


def parse(attrs: str) -> bool:
    '''Parse Block Attributes line:
       .class-names #id "css-properties" [html-attributes] block-options
    '''
    global classes, id, css, attributes, opts, ids
    if options.skipBlockAttributes():
        return True
    text = attrs
    text = utils.replaceInline(text, Expand(macros=True))

    # Kludge: The regexp is split in two to fix a catastrophic backtracking issue (without
    # this split the match was taking around 5 seconds!).
    # See https://unix.stackexchange.com/questions/419545/why-is-regular-expression-matching-so-slow
    # Split re after this asterisk     * <-- split regexp here.
    # r'^\\?\.((?:\s*[a-zA-Z][\w\-]*)+)*(?:\s*)?(#[a-zA-Z][\w\-]*\s*)?(?:\s*)?(?:"(.+?)")?(?:\s*)?(\[.+])?(?:\s*)?([+-][ \w+-]+)?$'
    # class-names = $1, id = $2, css-properties = $3, html-attributes = $4, block-options = $5'''
    r1 = re.compile(r'^\\?\.((?:\s*[a-zA-Z][\w\-]*)+)*')
    m1 = r1.match(text)
    if m1 is None:
        return False
    # Prefixed empty placeholder group () to maintain group match indexes in m2.
    r2 = re.compile(r'()(?:\s*)?(#[a-zA-Z][\w\-]*\s*)?(?:\s*)?(?:"(.+?)")?(?:\s*)?(\[.+])?(?:\s*)?([+-][ \w+-]+)?$')
    m2 = r2.match(text[m1.end():])
    if m2 is None:
        return False
    if m1[1]:
        # HTML element class names.
        classes += f' {m1[1].strip()}'
        classes = classes.strip()
    if m2[2]:
        # HTML element id.
        id = m2[2].strip()[1:]
    if m2[3]:
        # CSS properties.
        if css and not css.endswith(';'):
            css += ';'
        css += ' ' + m2[3].strip()
        css = css.strip()
    if m2[4] and not options.isSafeModeNz():
        # HTML attributes.
        attributes += ' ' + m2[4][1:- 1].strip()
        attributes = attributes.strip()
    if m2[5]:
        opts.parse(m2[5])
    return True


def injectHtmlAttributes(tag: str, consume: bool=True) -> str:
    '''Inject HTML attributes into the HTML `tag` and return result.
       Consume HTML attributes unless the 'tag' argument is blank.'''
    global classes, id, css, attributes, opts, ids
    if not tag:
        return tag
    result = tag
    attrs = ''
    if classes:
        match = re.compile(r'^(<[^>]*class=")(.*?)"', re.IGNORECASE).search(result)
        if match:
            # Inject class names into existing class attribute in first tag.
            result = result.replace(match[0], f'{match[1]}{classes} {match[2]}"', 1)
        else:
            attrs = f'class="{classes}"'
    if id:
        id = id.lower()
        has_id = re.compile(r'^<[^<]*id=".*?"', re.IGNORECASE).search(result)
        if has_id or id in ids:
            options.errorCallback(f"duplicate 'id' attribute: {id}")
        else:
            ids.insert(0, id)
        if not has_id:
            attrs += f' id="{id}"'
    if css:
        match = re.compile(r'^(<[^>]*style=")(.*?)"', re.IGNORECASE).search(result)
        if match:
            # Inject CSS styles into first style attribute in first tag.
            group2 = match[2].strip()
            if not group2.endswith(';'):
                group2 += ';'
            result = result.replace(match[0], f'{match[1]}{group2} {css}"', 1)
        else:
            attrs += f' style="{css}"'
    if attributes:
        attrs += f' {attributes}'
    attrs = attrs.strip()
    if attrs:
        match = re.compile(r'^<([a-z]+|h[1-6])(?=[ >])', re.IGNORECASE).search(result)
        if match:
            # Inject attributes after tag name.
            before = result[:len(match[0])]
            after = result[len(match[0]):]
            result = before + ' ' + attrs + after
    if consume:
        # Consume the attributes.
        classes = ''
        id = ''
        css = ''
        attributes = ''
    return result


def slugify(text: str) -> str:
    slug = re.sub(r'\W+', '-', text)  # Replace non-alphanumeric characters with dashes.
    slug = re.sub(r'-+', '-', slug)  # Replace multiple dashes with single dash.
    slug = re.sub(r'(^-)|(-$)', '', slug)  # Trim leading and trailing dashes.
    slug = slug.lower()
    if not slug:
        slug = 'x'
    if slug in ids:
        # Another element already has that id.
        i = 2
        while f'{slug}-{i}' in ids:
            i += 1
        slug += f'-{i}'
    return slug



import io
import os
import re
import sys
from typing import List, Optional

import rimu

import rimuc

VERSION = '11.3.0'
NAME = 'rimupy'
HOME_DIR = os.path.expanduser('~')
RIMURC = os.path.join(HOME_DIR, '.rimurc')


def die(message: str = '') -> None:
    if message != '':
        sys.stderr.write(message+'\n')
    raise SystemExit(1)


def readResource(name: str) -> str:
    if name not in rimuc.resources:
        die(f'missing resource: {name}')
    return rimuc.resources[name]


def main() -> None:
    '''Process sys.argv command-line.'''
    RESOURCE_TAG = 'resource:'  # Placeholder tag for resource files.
    PREPEND_TAG = '--prepend options'    # Placeholder tag for prepend options.
    STDIN = '-'

    # Command option values.
    safe_mode: Optional[int] = None
    html_replacement: Optional[str] = None
    layout: str = ''
    no_rimurc: bool = False
    prepend_files: List[str] = []
    pass_through: bool = False

    def popArg(arg: str) -> str:
        if len(args) == 0:
            die(f'missing {arg} option value')
        return args.pop(0)

    # Parse command-line options.
    prepend: str = ''
    outfile: str = ''
    args = sys.argv[1:]  # Drop script name.
    while len(args) > 0:
        arg: str = args.pop(0)
        if arg in ['--help', '-h']:
            manpage = readResource('manpage.txt')
            manpage = re.sub(r'\brimuc\b', NAME, manpage)
            print('\n' + manpage)
            return
        elif arg in ['--version']:
            print(VERSION)
            return
        elif arg in ['--lint', '-l']:   # Deprecated in Rimu 10.0.0
            pass
        elif arg in ['--output', '-o']:
            outfile = popArg(arg)
        elif arg in ['--pass']:
            pass_through = True
        elif arg in ['--prepend', '-p']:
            prepend += popArg(arg) + '\n'
        elif arg in ['--prepend-file']:
            prepend_file: str = popArg(arg)
            prepend_files.append(prepend_file)
        elif arg in ['--no-rimurc']:
            no_rimurc = True
        elif arg in ['--safe-mode', '--safeMode']:  # --safeMode deprecated in Rimu 7.1.0
            safe_mode = int(popArg(arg))
            if safe_mode < 0 or safe_mode > 15:
                die(f'illegal --safe-mode option value: {safe_mode}')
        elif arg in ['--html-replacement', '--htmlReplacement']:  # --htmlReplacement deprecated in Rimu 7.1.0
            html_replacement = popArg(arg)
        elif arg in [  # Styling macro definitions shortcut options.
            '--highlightjs',
            '--mathjax',
            '--section-numbers',
            '--theme',
            '--title',
            '--lang',
            '--toc',             # --toc deprecated in Rimu 8.0.0
            '--no-toc',
            '--sidebar-toc',     # --sidebar-toc deprecated in Rimu 10.0.0
            '--dropdown-toc',    # --dropdown-toc deprecated in Rimu 10.0.0
            '--custom-toc',
            '--header-ids',
                '--header-links']:
            macro_value = popArg(arg) if arg in ['--lang', '--title', '--theme'] else 'true'
            prepend += f"{{{arg}}}='{macro_value}'\n"
        elif arg in ['--layout', '--styled-name']:  # --styled-name' deprecated in Rimu 10.0.0
            layout = popArg(arg)
            if layout not in ['classic', 'flex', 'plain', 'sequel', 'v8']:
                # NOTE: Imported layouts are not supported.
                die(f"illegal --layout: {layout}")
            prepend += "{--header-ids}='true'\n"
        elif arg in ['--styled', '-s']:
            prepend += "{--header-ids}='true'\n"
            prepend += "{--no-toc}='true'\n"
            layout = 'sequel'
        else:
            args.insert(0, arg)  # Contains source file names.
            break
    # process.argv contains the list of source files.
    files = list(args)
    if len(files) == 0:
        files.append(STDIN)
    elif len(files) == 1 and layout != '' and files[0] != STDIN and len(outfile) == 0:
        # Use the source file name with .html extension for the output file.
        outfile = os.path.splitext(files[0])[0] + '.html'
    if layout != '':
        # Envelope source files with header and footer.
        files.insert(0, f'{RESOURCE_TAG}{layout}-header.rmu')
        files.append(f'{RESOURCE_TAG}{layout}-footer.rmu')
    # Prepend $HOME/.rimurc file if it exists.
    if not no_rimurc and os.path.isfile(RIMURC):
        prepend_files.insert(0, RIMURC)
    if prepend:
        prepend_files.append(PREPEND_TAG)
    files = prepend_files + files
    # Convert Rimu source files to HTML.
    output = ''
    errors = 0
    options = rimu.RenderOptions()
    if html_replacement is not None:
        options.htmlReplacement = html_replacement
    for infile in files:
        source = ''
        options.safeMode = safe_mode
        ext = ''
        if infile.startswith(RESOURCE_TAG):
            infile = infile[len(RESOURCE_TAG):]
            source = readResource(infile)
            options.safeMode = 0  # Resources are trusted.
        elif infile == STDIN:
            source = sys.stdin.read()
        elif infile == PREPEND_TAG:
            source = prepend
            options.safeMode = 0  # --prepend options are trusted.
        else:
            if not os.path.isfile(infile):
                die('source file does not exist: ' + infile)
            try:
                with open(infile) as f:
                    source = f.read()
            except:
                die('source file permission denied: ' + infile)
            if infile in prepend_files:
                # Prepended and ~/.rimurc files are trusted.
                options.safeMode = 0
            ext = os.path.splitext(infile)[1]
        # Skip .html and pass-through inputs.
        if not (ext == '.html' or (pass_through and infile == STDIN)):
            def callback(message: rimu.CallbackMessage) -> None:
                nonlocal errors
                msg = f'{message.type}: {"/dev/stdin" if infile == STDIN else infile}: {message.text}'
                if len(msg) > 120:
                    msg = msg[: 117] + '...'
                sys.stderr.write(msg + '\n')
                if message.type == 'error':
                    errors += 1
            options.callback = callback
            source = rimu.render(source, options)
        source = source.strip()
        if source:
            output += source + '\n'
    output = output.strip()
    if len(outfile) == 0 or outfile == '-':
        sys.stdout.write(output)
    else:
        with open(outfile, 'w') as f:
            f.write(output)
    if errors > 0:
        die()



from typing import Any, Callable, Optional

from rimu import api, utils


Callback = Optional[Callable[['CallbackMessage'], None]]


# Global option values.
safeMode: int = -1  # Trigger API initialization.
htmlReplacement: str
callback: Callback = None


class RenderOptions:
    '''Rimu render API options.'''
    safeMode: Optional[int]
    htmlReplacement: Optional[str]
    reset: Optional[bool]
    callback: Callback

    def __init__(self,
                 safeMode: int = None,
                 htmlReplacement: str = None,
                 reset: Any = None,
                 callback: Callback = None,
                 ):
        self.safeMode = safeMode
        self.htmlReplacement = htmlReplacement
        self.reset = reset
        self.callback = callback


class CallbackMessage:
    type: str
    text: str

    def __init__(self, type: str, text: str):
        self.type = type
        self.text = text


def init() -> None:
    '''Initialize API render options.'''
    global safeMode, htmlReplacement, callback
    safeMode = 0
    htmlReplacement = '<mark>replaced HTML</mark>'
    callback = None


def isSafeModeNz() -> bool:
    '''Return true if safeMode is non-zero.'''
    return safeMode != 0


def getSafeMode() -> int:
    return safeMode


def skipMacroDefs() -> bool:
    '''Return true if Macro Definitions are ignored.'''
    return safeMode != 0 and (safeMode & 0x8) == 0


def skipBlockAttributes() -> bool:
    '''Return true if Block Attribute elements are ignored.'''
    return (safeMode & 0x4) != 0


def updateFrom(options: RenderOptions) -> None:
    ''' Update specified (non-null) options.'''
    global safeMode, htmlReplacement, callback
    # Install callback first to ensure option errors are logged.
    if callback is not None:
        callback = options.callback
    setOption('reset', options.reset)  # Reset takes priority.
    # Install callback again in case it has been reset.
    if options.callback is not None:
        callback = options.callback
    if options.safeMode is not None:
        setOption('safeMode', str(options.safeMode))
    if options.htmlReplacement is not None:
        setOption('htmlReplacement', options.htmlReplacement)


def setOption(name: str, value: Any) -> None:
    '''Set named option value.'''
    global safeMode, htmlReplacement, callback
    if name == 'safeMode':
        n = 0
        try:
            n = int(value)
        except:
            errorCallback('illegal safeMode API option value: ' + str(value))
        if n < 0 or n > 15:
            errorCallback('illegal safeMode API option value: ' + str(value))
        else:
            safeMode = n
    elif name == 'reset':
        if value is None or value == False or value == 'false':
            return
        elif value == True or value == 'true':
            api.init()
        else:
            errorCallback('illegal reset API option value: ' + str(value))
    elif name == 'htmlReplacement':
        htmlReplacement = str(value)
    else:
        errorCallback('illegal API option name: ' + name)


def htmlSafeModeFilter(html: str) -> str:
    '''Filter HTML based on current safeMode.'''
    n = safeMode & 0x3
    if n == 0:    # Raw HTML (default behavior).
        return html
    elif n == 1:  # Drop HTML.
        return ''
    elif n == 2:  # Replace HTML with 'htmlReplacement' option string.
        return htmlReplacement
    elif n == 3:  # Render HTML as text.
        return utils.replaceSpecialChars(html)
    else:
        return ''


def errorCallback(message: str) -> None:
    if callback is not None:
        callback(CallbackMessage('error', message))


def panic(message: str) -> None:
    '''Called when an unexpected program error occurs.'''
    msg = 'panic: ' + message
    print(msg)
    errorCallback(msg)



from rimu import (blockattributes, delimitedblocks, io, lineblocks, lists,
                  macros, options, quotes, replacements)


def init() -> None:
    blockattributes.init()
    options.init()
    delimitedblocks.init()
    macros.init()
    quotes.init()
    replacements.init()


def render(source: str) -> str:
    reader = io.Reader(source)
    writer = io.Writer()
    while not reader.eof():
        reader.skipBlankLines()
        if reader.eof():
            break
        if lineblocks.render(reader, writer):
            continue
        if lists.render(reader, writer):
            continue
        if delimitedblocks.render(reader, writer):
            continue
        # This code should never be executed (normal paragraphs should match anything).
        options.panic('no matching delimited block found')
    return writer.toString()



# Generated automatically from resource files. Do not edit.
from typing import Dict
resources: Dict[str, str] = {
    'sequel-footer.rmu': r'''/*
  Used by rimuc `--layout sequel` option.
*/

// Close main and article divs.
</div>
</div>

{--highlightjs=}.+skip
{--highlightjs-scripts}

{--mathjax!}{--mathjax-scripts}

{--no-toc!}.+skip
<script>
window.onload = function() {
  document.getElementById('sidebar').appendChild(document.getElementById('toc')); // Ensure custom TOC is child of sidebar.
  var headings = [].slice.call(document.body.querySelectorAll('#article > h1, #article > h2, #article > h3'));
  headings.forEach(function(heading) {
{--header-links!}    setHeaderLink(heading);
    appendTocEntry(heading);
  });
  if (isSmallScreen()) {
    toggleToc();  // Hide TOC.
  }
}
</script>

{--header-links=}.+skip
<script>
function setHeaderLink(heading) {
  var id = heading.getAttribute('id');
  if (id) {
    var link = document.createElement('a');
    link.classList.add('header-link');
    link.setAttribute('href', '#' + id);
    heading.appendChild(link);
  }
}
</script>

{--no-toc!}.+skip
<script>
function appendTocEntry(heading) {
  var id = heading.getAttribute('id');
  if (heading.classList.contains('no-auto-toc')) {
    return;
  }
  var container = document.getElementById('auto-toc');
  if (container === null) {
    return;
  }
  var tocLink = document.createElement('a');
  tocLink.setAttribute('href', '#' + id);
  tocLink.textContent = heading.textContent;
  var tocEntry = document.createElement('div');
  tocEntry.setAttribute('class', heading.tagName.toLowerCase());
  tocEntry.appendChild(tocLink);
  container.appendChild(tocEntry);
}
</script>

{--no-toc!}.+skip
<script>
  document.onclick = function (event) {
    if (event.target.matches('#toc-button') || event.target.matches('#toc a') && isSmallScreen()) {
{--!} Toggle TOC if TOC button or small-screen TOC link is clicked.
      toggleToc();
    }
  }
  function toggleToc() {
    document.body.classList.toggle('hide-toc');
  }
  function isSmallScreen() {
    return window.matchMedia('{--small-screen}').matches;
  }
  // matches() polyfill for old browsers.
  if (!Element.prototype.matches) {
    var p = Element.prototype;
    if (p.webkitMatchesSelector) // Chrome <34, SF<7.1, iOS<8
      p.matches = p.webkitMatchesSelector;
    if (p.msMatchesSelector) // IE9/10/11 & Edge
      p.matches = p.msMatchesSelector;
    if (p.mozMatchesSelector) // FF<34
      p.matches = p.mozMatchesSelector;
  }
</script>

</body>
</html>''',
    'plain-footer.rmu': r'''</body>
</html>''',
    'manpage.txt': r'''NAME
  rimuc - convert Rimu source to HTML

SYNOPSIS
  rimuc [OPTIONS...] [FILES...]

DESCRIPTION
  Reads Rimu source markup from stdin, converts them to HTML
  then writes the HTML to stdout. If FILES are specified
  the Rimu source is read from FILES. The contents of files
  with an .html extension are passed directly to the output.
  An input file named '-' is read from stdin.

  If a file named .rimurc exists in the user's home directory
  then its contents is processed (with --safe-mode 0).
  This behavior can be disabled with the --no-rimurc option.

  Inputs are processed in the following order: .rimurc file then
  --prepend-file option files then --prepend option source and
  finally FILES...

OPTIONS
  -h, --help
    Display help message.

  --html-replacement TEXT
    Embedded HTML is replaced by TEXT when --safe-mode is set to 2.
    Defaults to '<mark>replaced HTML</mark>'.

  --layout LAYOUT
    Generate a styled HTML document. rimuc includes the
    following built-in document layouts:

    'classic': Desktop-centric layout.
    'flex':    Flexbox mobile layout (experimental).
    'plain':   Unstyled HTML layout.
    'sequel':  Responsive cross-device layout.

    If only one source file is specified and the --output
    option is not specified then the output is written to a
    same-named file with an .html extension.
    This option enables --header-ids.

  -s, --styled
    Style output using default layout.
    Shortcut for '--layout sequel --header-ids --no-toc'

  -o, --output OUTFILE
    Write output to file OUTFILE instead of stdout.
    If OUTFILE is a hyphen '-' write to stdout.

  --pass
    Pass the stdin input verbatim to the output.

  -p, --prepend SOURCE
    Process the Rimu SOURCE text (immediately after --prepend-file
    option files). Rendered with --safe-mode 0. This option can be
    specified multiple times.

  --prepend-file PREPEND_FILE
    Process the PREPEND_FILE contents (immediately after .rimurc file).
    Rendered with --safe-mode 0. This option can be specified
    multiple times.

  --no-rimurc
    Do not process .rimurc from the user's home directory.

  --safe-mode NUMBER
    Non-zero safe modes ignore: Definition elements; API option elements;
    HTML attributes in Block Attributes elements.
    Also specifies how to process HTML elements:

    --safe-mode 0 renders HTML (default).
    --safe-mode 1 ignores HTML.
    --safe-mode 2 replaces HTML with --html-replacement option value.
    --safe-mode 3 renders HTML as text.

    Add 4 to --safe-mode to ignore Block Attribute elements.
    Add 8 to --safe-mode to allow Macro Definitions.

  --theme THEME, --lang LANG, --title TITLE, --highlightjs, --mathjax,
  --no-toc, --custom-toc, --section-numbers, --header-ids, --header-links
    Shortcuts for the following prepended macro definitions:

    --prepend "{--custom-toc}='true'"
    --prepend "{--header-ids}='true'"
    --prepend "{--header-links}='true'"
    --prepend "{--highlightjs}='true'"
    --prepend "{--lang}='LANG'"
    --prepend "{--mathjax}='true'"
    --prepend "{--no-toc}='true'"
    --prepend "{--section-numbers}='true'"
    --prepend "{--theme}='THEME'"
    --prepend "{--title}='TITLE'"

  --version
    Print version number.

LAYOUT OPTIONS
  The following options are available when the --layout option
  is used:

  Option             Description
  _______________________________________________________________
  --custom-toc       Set to a non-blank value if a custom table
                     of contents is used.
  --header-links     Set to a non-blank value to generate h2 and
                     h3 header header links.
  --highlightjs      Set to non-blank value to enable syntax
                     highlighting with Highlight.js.
  --lang             HTML document language attribute value.
  --mathjax          Set to a non-blank value to enable MathJax.
  --no-toc           Set to a non-blank value to suppress table
                     of contents generation.
  --section-numbers  Apply h2 and h3 section numbering.
  --theme            Styling theme. Theme names:
                     'legend', 'graystone', 'vintage'.
  --title            HTML document title.
  _______________________________________________________________
  These options are translated by rimuc to corresponding layout
  macro definitions using the --prepend option.

LAYOUT CLASSES
  The following CSS classes can be used to style Rimu block
  elements in conjunction with the --layout option:

  CSS class        Description
  _______________________________________________________________
  align-center     Align element content center.
  align-left       Align element content left.
  align-right      Align element content right.
  bordered         Add borders to table element.
  cite             Quote and verse attribution.
  dl-horizontal    Format labeled lists horizontally.
  dl-numbered      Number labeled list items.
  dl-counter       Prepend dl item counter to element content.
  ol-counter       Prepend ol item counter to element content.
  ul-counter       Prepend ul item counter to element content.
  no-auto-toc      Exclude heading from table of contents.
  no-page-break    Avoid page break inside the element.
  no-print         Do not print element.
  page-break       Force a page break before the element.
  preserve-breaks  Honor line breaks in element content.
  sidebar          Paragraph and division block style.
  verse            Paragraph and division block style.
  important        Paragraph and division block style.
  note             Paragraph and division block style.
  tip              Paragraph and division block style.
  warning          Paragraph and division block style.
  _______________________________________________________________

PREDEFINED MACROS
  Macro name         Description
  _______________________________________________________________
  --                 Blank macro (empty string).
                     The Blank macro cannot be redefined.
  --header-ids       Set to a non-blank value to generate h1, h2
                     and h3 header id attributes.
  _______________________________________________________________''',
    'classic-header.rmu': r'''/*
  Used by rimuc `--layout classic` option.
*/

/*
  Default values for styling themes and content options.
*/

{--section-numbers?} = ''
{--lang?} = ''
{--title?} = '&nbsp;'
{--theme?} = ''
{--no-toc?} = ''
{--custom-toc?} = ''
{--header-links?} = ''
{--!} The min-width: 1px clause stops page load transitions in IE11 and Edge (https://stackoverflow.com/a/25850649).
{--small-screen?} = '(min-width: 1px) and (max-width: 800px)'
{--meta?} = '<meta charset="UTF-8">
  {--!} Make old IE versions use the latest rendering engine.
  <meta http-equiv="X-UA-Compatible" content="IE=edge">
  <meta name="viewport" content="width=device-width, initial-scale=1, minimum-scale=1">'
// Additional <head> child elements.
{--head?} = ''

{--highlightjs?} = ''
{--highlightjs-css} = '<link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.5.0/styles/default.min.css">'
{--highlightjs-scripts} = '<script src="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.5.0/highlight.min.js"></script>
<script>
hljs.configure({languages: []});
hljs.highlightAll();
</script>'

{--mathjax?} = ''
{--mathjax-scripts} = '<script async src="https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.2/MathJax.js?config=TeX-MML-AM_CHTML"></script>'

// List item CSS counters.
{--dl-counter} = '<span class="dl-counter"></span>'
{--ol-counter} = '<span class="ol-counter"></span>'
{--ul-counter} = '<span class="ul-counter"></span>'

// Classic layout specific.
{--dropdown-toc?} = ''

/*
  Legend theme.
*/
{--theme!.*\blegend\b.*}.+skip
..
{--sans-font?} = 'Helvetica, Arial, sans-serif'
{--serif-font?} = 'Georgia, Times, serif'
{--mono-font?} = 'Consolas, "Andale Mono", Monaco, monospace'
{--mono-size?} = '90%'
{--text-color?} = '#333333'
{--border-color?} = '#dddddd'
{--primary-color?} = '#34495e'
{--primary-background?} = 'white'
{--code-background?} = '#f8f8f8'
{--sidebar-background?} = '#ffffee'
{--link-color?} = '#428bca'
{--border-radius?} = '0'
{--max-text-width?} = '50rem'
..

/*
  Vintage theme.
*/
{--theme!.*\bvintage\b.*}.+skip
// Default to this theme if none of the built-in themes is specified.
{--theme!.*\b(legend|vintage|graystone)\b.*}.-skip
..
{--sans-font?} = 'Helvetica, Arial, sans-serif'
{--serif-font?} = 'Georgia, Times, serif'
{--mono-font?} = 'Consolas, "Andale Mono", Monaco, monospace'
{--mono-size?} = '90%'
{--text-color?} = '#333333'
{--border-color?} = '#dddddd'
{--primary-color?} = '#527bbd'
{--primary-background?} = 'white'
{--code-background?} = '#f8f8f8'
{--sidebar-background?} = '#ffffee'
{--link-color?} = '#527bbd'
{--border-radius?} = '4px'
{--max-text-width?} = '50rem'
..

/*
  Graystone theme.
*/
{--theme!.*\bgraystone\b.*}.+skip
..
{--sans-font?} = 'Helvetica, Arial, sans-serif'
{--serif-font?} = 'Georgia, Times, serif'
{--mono-font?} = 'Consolas, "Andale Mono", Monaco, monospace'
{--mono-size?} = '90%'
{--text-color?} = '#333333'
{--border-color?} = '#dddddd'
{--primary-color?} = '#888'
{--primary-background?} = 'white'
{--code-background?} = '#f8f8f8'
{--sidebar-background?} = '#ffffee'
{--link-color?} = '#888'
{--border-radius?} = '0'
{--max-text-width?} = '50rem'
..

/*
  Private layout parameters.
*/
// Accomodate the new default TOC generation and --no-toc option.
// DEPRECATED --sidebar-toc: Used internally, external definition tolerated for backward comaptibility.
{--dropdown-toc!}.+skip
..
{--sidebar-toc?} = 'yes'
..

{--dropdown-toc=}.+skip
..
{--sidebar-toc?} = ''
..

// DEPRECATED --toc: If --toc is non-blank make --sidebar-toc non-blank.
{--toc?} = ''
{--sidebar-toc} = '{--toc}{--sidebar-toc}'

// Ensures no TOC.
{--no-toc=}.+skip
..
{--dropdown-toc} = ''
{--sidebar-toc} = ''
..


/*
  HTML header
*/
<!DOCTYPE HTML>
{--lang=}<html>
{--lang!}<html lang="{--lang}">
<head>
{--meta}
<title>{--title}</title>
{--highlightjs!}{--highlightjs-css}


/*
  Layout independent styles
*/
<style>
  :root {
    font-size: 16px;
    line-height: 20px;
    color: {--text-color};
    background-color: {--primary-background};
    font-family: {--sans-font};
  }
  * {
    margin: 0;
    padding: 0;
    white-space: inherit;
    box-sizing: border-box;
  }
  h1, h2, h3, h4, h5, h6 {
    margin: 10px 0;
    color: {--primary-color};
  }
  h1 { font-size: 2.1rem; line-height: 40px;}
  h2 { font-size: 1.5rem; }
  h3 { font-size: 1.4rem; }
  h4 { font-size: 1.1rem; }
  h5 { font-size: 1.0rem; }
  h6 { font-size: 0.9rem; }
  h2, h3, h4, h5, h6 {
    line-height: normal;
    margin-top: 1.0rem;
    margin-bottom: 0.2rem;
  }
  h2 {
    margin-top: 1.5rem;
    border-bottom: 1px solid {--border-color};
  }
  img {
    max-width: 100%;
  }
  table {
    border-collapse: collapse;
    table-layout: fixed;
    width: 100%;
  }
  td, th {
    text-align: left;
    vertical-align: top;
  }
  {--!} tbody necessary because: https://stackoverflow.com/questions/7490364/why-do-browsers-still-inject-tbody-in-html5
  table.bordered > tbody > tr > th,
  table.bordered > tbody > tr > td {
    border: 1px solid {--border-color};
  }
  a, a:hover {
    color: {--primary-color};
  }
  a {
    text-decoration: none;
  }
  a:hover {
    text-decoration: underline;
  }
  p, pre, li, dt, dd, blockquote p {
    line-height: 1.45;
    margin-top: 0.5rem;
    margin-bottom: 0.5rem;
  }
  ul, ol {
    padding: 0;
    margin: 0 0 10px 25px;
  }
  dt {
    margin-bottom: 0;
    font-style: italic;
  }
  dd {
    margin-left: 1.0rem;
    margin-top: 0;
  }
  blockquote {
    padding: 0 0 0 15px;
    border: none;
    border-left: 5px solid #eeeeee;
    font-family: {--serif-font};
    width: 80%;
    margin: 1.5em 0;
    color: #383838;
  }
  .cite {
    color: #777;
    padding: 5px 0;
  }
  .cite::before {
    content: "\2014 \2009";
  }
  code {
    font-family: {--mono-font};
    {--!} Monspace fonts are relatively oversized.
    font-size: {--mono-size};
    background-color: {--code-background};
  }
  pre {
    font-family: {--mono-font};
    line-height: normal;
    white-space: pre-wrap;
    background-color: {--code-background};
    border: 1px solid {--border-color};
    border-radius: {--border-radius};
    padding: 0;
  }
  .light-background {
    background-color: {--code-background};
    border: none;
    white-space: pre-wrap;
  }
  .dl-horizontal > dd {
    margin-top: 1.0rem;
  }
  pre > code {
    background-color: inherit;
    display: block;
    padding: 0.5em !important;
    border-radius: {--border-radius};
  }
  pre span {
    {--!} highlight.js tweak.
    opacity: 1 !important;
  }
  {--!} Rimu classes.
  {--!} Apply verse class to Normal Paragraphs and Division blocks.
  .verse {
    margin: 1.5em 20px;
  }
  div.verse p, p.verse {
    font-family: {--serif-font};
    white-space: pre-wrap;
    margin-top: 0.75rem;
    margin-bottom: 0.75rem;
  }
  {--!} Apply sidebar and admonition classes to Normal Paragraphs and Division blocks.
  .sidebar,
  .important,
  .note,
  .tip,
  .warning {
    margin: 1.5em 0;
    padding: 10px;
    border-radius: {--border-radius};
  }
  div.sidebar *:first-child,
  div.important *:first-child,
  div.note *:first-child,
  div.tip *:first-child,
  div.warning *:first-child {
    margin-top: 0.2rem !important;
  }
  .note::before,
  .important::before,
  .tip::before,
  .warning::before {
    font-weight: bold;
  }
  .notitle::before {
    content: "" !important;
  }
  .sidebar {
    background: #ffffee;
    border: 1px solid {--border-color};
  }
  .note {
    background-color: #f0f7fb;
    border-left: solid 4px #3498db;
  }
  div.note::before {
    content: "Note";
  }
  p.note::before {
    content: "Note: ";
  }
  .tip {
    background-color: #e7f6ef;
    border-left: solid 4px #32c875;
  }
  div.tip::before {
    content: "Tip";
  }
  p.tip::before {
    content: "Tip: ";
  }
  .warning {
    background-color: #fdf7f2;
    border-left: solid 4px #d1534a;
  }
  div.warning::before {
    content: "Warning";
  }
  p.warning::before {
    content: "Warning: ";
  }
  .important {
    background-color: #fffbea;
    border-left: solid 4px #eec51c;
  }
  div.important::before {
    content: "Important";
  }
  p.important::before {
    content: "Important: ";
  }
  {--!} Force page break before the element.
  .page-break {
    page-break-before: always;
  }
  {--!} Avoid page breaks inside the element.
  .no-page-break {
    page-break-inside: avoid;
  }
  {--!} Text block alignment classes.
  .align-left {
    text-align: left;
  }
  .align-center {
    text-align: center;
  }
  .align-right {
    text-align: right;
  }
  {--!} Preserve line breaks.
  .preserve-breaks {
    white-space: pre;
  }
  {--!} DEPRECATED: Use `preserve-breaks` instead.
  .line-breaks {
    white-space: pre-wrap;
  }
  {--!} Horizontal labeled list.
  .dl-horizontal:before, .dl-horizontal:after {
    display: table;
    content: "";
    line-height: 0;
  }
  .dl-horizontal:after {
    clear: both;
  }
  .dl-horizontal > dt {
    float: left;
    clear: left;
    overflow: hidden;
    text-overflow: ellipsis;
    white-space: nowrap;
    text-align: left;
    margin-top: 1.0rem;
    width: 25%;
  }
  .dl-horizontal > dd {
    margin-left: 25%;
    padding-left: 1.0rem;
  }
  {--!} List item counters.
  dl {
    counter-reset: dl-counter;
  }
  dl > dt {
    counter-increment: dl-counter;
  }
  ol {
    counter-reset: ol-counter;
  }
  ol > li {
    counter-increment: ol-counter;
  }
  ul {
    counter-reset: ul-counter;
  }
  ul > li {
    counter-increment: ul-counter;
  }
  {--!} Prefix list counter to counter class element content.
  .dl-counter:before {
    content: counter(dl-counter) " ";
  }
  .ol-counter:before {
    content: counter(ol-counter) " ";
  }
  .ul-counter:before {
    content: counter(ul-counter) " ";
  }
  {--!} Number labeled list items.
  .dl-numbered > dt:before {
    content: counter(dl-counter) ". ";
  }
  {--!} Printing.
  @media print {
    .no-print, .no-print * {
      display: none !important;
    }
    a:hover {
      text-decoration: none !important;
    }
  }
</style>

{--section-numbers=}.+skip
<style>
  /* Section numbers. */
  body,h1 { counter-reset: h2-counter; }
  h2      { counter-reset: h3-counter; }
  #article > h2:before {
    content: counter(h2-counter) ". ";
    counter-increment: h2-counter;
  }
  #article > h3:before {
    content: counter(h2-counter) "." counter(h3-counter) ". ";
    counter-increment: h3-counter;
  }
</style>

// Common to --sidebar-toc and --dropdown-toc.
{--no-toc!}.+skip
<style>
  #toc {
    padding-left: 1rem;
  }
  #toc h2 {
    font-size: 125%;
    margin-top:1.0rem;
    margin-left: 0;
  }
  #toc .h1 {
    font-size: 110%;
    font-weight: bold;
    margin-top: 0.5rem;
    margin-bottom: 0.4rem;
  }
  #toc .h2 {
    margin-top: 0.4rem;
  }
  #toc .h3 {
    margin-left: 1.5rem;
    font-size: 90%;
  }
  #toc div[class^="h"]:nth-child(even) {
    background-color: #f8f8f8;
  }
  #auto-toc {
    width: 100%;
  }
</style>

{--header-links=}.+skip
<style>
  .header-link {
    text-decoration: none;
    color: lightgray;
    visibility: hidden;
  }
  .header-link:hover {
    text-decoration: none;
    color: gray;
  }
  .header-link::before {
    content: " \00b6";
  }
  h2:hover > .header-link, h3:hover > .header-link {
    visibility: visible;
  }
  @media print {
    .header-link {
      display: none !important;
    }
  }
</style>


/*
  Layout styles
*/
<style>
  #article {
    padding-left: 2rem;
    padding-right: 1rem;
    max-width: {--max-text-width};
  }
</style>

{--sidebar-toc=}.+skip
<style>
  #article {
    margin-left: 20rem;
  }
  #toc {
    position: fixed;
    top: 0;
    left: 0;
    bottom: 0;
    width: 20rem;
    border-right: 1px solid #cccccc;
    overflow-y: auto;
    overflow-x: hidden;
    box-shadow: 0 0 3px rgba(0, 0, 0, 0.35);
  }
  @media screen and {--small-screen} {
    {--!}  Sidebar TOC hidden as it does not work on small screens.
    #article {
      margin-left: 0;
    }
    #toc {
      display: none !important;
    }
  }
</style>

{--dropdown-toc=}.+skip
<style>
  .show-toc #toc {
    display: block !important;
  }
  #toc-button {
    position: fixed;
    top: 22px;
    left: 15px;
    z-index: 1;
    cursor: pointer;
    color: silver;
    font-size: 3.2rem;
  }
  #toc {
    display: none;
    position: fixed;
    top: 55px;
    left: 17px;
    right: 10%;
    z-index: 1;
    max-width: 30rem;
    max-height: 80%;
    overflow-y: auto;
    background-color: {--primary-background};
    box-shadow: 0 0 10px rgba(0, 0, 0, 0.35);
  }
  #article {
    margin-left: 2rem;
  }
</style>

// Format (non-layout) related styles.
<style>
  @media screen and {--small-screen} {
    body {
      font-size: 20px;
    }
    * {
      overflow-wrap: break-word;
    }
  }
  @media print {
    #article {
      margin-left: 0;
      max-width: 100%;
    }
  }
</style>

/*
  Theme related styles
*/
{--theme!.*\bgraystone\b.*}.+skip
<style>
  h1, h2 {
    text-transform: uppercase;
  }
</style>

{--head}

</head>
<body>

// Include dropdown TOC button.
{--dropdown-toc=}.+skip
<div id="toc-button" class="no-print">&#8801;</div>

// Include TOC unless a custom TOC is specified.
.+skip
{--no-toc=}.-skip
{--custom-toc!}.+skip
<div id="toc" class="no-print">
  <div id="auto-toc"></div>
</div>

<div id="article">''',
    'v8-footer.rmu': r'''/*
  Used by rimuc `--layout v8` option.
  DEPRECATED: This layout is no longer maintained, for Rimu version 8 compatibility.
*/

// Close contents div.
</div>

{--highlightjs!}<script src="https://yandex.st/highlightjs/7.3/highlight.min.js"></script><script>hljs.initHighlightingOnLoad();</script>

{--mathjax!}<script type="text/javascript" async src="https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.2/MathJax.js?config=TeX-MML-AM_CHTML"></script>

<script>
window.onload = function() {
  var headings = [].slice.call(document.body.querySelectorAll('#contents > h1, #contents > h2, #contents > h3'));
  headings.forEach(function(heading, index) {
    var title = heading.textContent;
    var id = heading.getAttribute('id');
    if (!id) {
      id = slugify(title);
      heading.setAttribute('id', id);
    }
    if (index === 0 && heading.tagName === 'H1') {
      id = ''; // Go to top of page.
    }
{--sidebar-toc!}    appendTocEntry(heading, id);
{--dropdown-toc!}    appendTocEntry(heading, id);
  });
}
function slugify(text) {
  var slug = text.replace(/\s+/g, '-') // Replace spaces with dashes.
      .replace(/[^\w-]/g, '')          // Retain alphanumeric, '-' and '_' characters.
      .toLowerCase()
  if (!slug) slug = 'x';
  if (document.getElementById(slug)) { // Another element already has that id.
    var i = 2, prefix = slug;
    while (document.getElementById(slug = prefix + '-' + i++)) {}
  }
  return slug;
}
</script>

.+skip
{--sidebar-toc!}.-skip
{--dropdown-toc!}.-skip
<script>
function appendTocEntry(heading, id) {
  if (heading.classList.contains('no-auto-toc')) {
    return;
  }
  var container = document.getElementById('auto-toc');
  if (container === null) {
    return;
  }
  var tocLink = document.createElement('a');
  tocLink.setAttribute('href', '#' + id);
  tocLink.textContent = heading.textContent;
  var tocEntry = document.createElement('div');
  tocEntry.setAttribute('class', heading.tagName.toLowerCase());
  tocEntry.appendChild(tocLink);
  container.appendChild(tocEntry);
}
</script>

{--dropdown-toc=}.+skip
<script>
function toggleToc() {
    document.getElementById("toc").classList.toggle('toc-visible');
}
window.onclick = function(event) {
  if (!Element.prototype.matches) {
    // matches() polyfill for old browsers.
    var p = Element.prototype;
    if (p.webkitMatchesSelector) // Chrome <34, SF<7.1, iOS<8
      p.matches = p.webkitMatchesSelector;
    if (p.msMatchesSelector) // IE9/10/11 & Edge
      p.matches = p.msMatchesSelector;
    if (p.mozMatchesSelector) // FF<34
      p.matches = p.mozMatchesSelector;
  }
  if (!event.target.matches('#toc-button, #toc, #toc :not(a)')) {
    // Hide TOC if clicked outside TOC or on TOC link.
    var toc = document.getElementById('toc');
    if (toc.classList.contains('toc-visible')) {
      toc.classList.remove('toc-visible');
    }
  }
}
</script>

</body>
</html>''',
    'classic-footer.rmu': r'''/*
  Used by rimuc.js --styled option.
*/

// Close article div.
</div>

{--highlightjs=}.+skip
{--highlightjs-scripts}

{--mathjax!}{--mathjax-scripts}

.+skip
{--no-toc=}.-skip
{--header-links!}.-skip
<script>
window.onload = function() {
  var headings = [].slice.call(document.body.querySelectorAll('#article > h1, #article > h2, #article > h3'));
  headings.forEach(function(heading) {
{--header-links!}    setHeaderLink(heading);
{--no-toc=}    appendTocEntry(heading);
  });
}
</script>

{--header-links=}.+skip
<script>
function setHeaderLink(heading) {
  var id = heading.getAttribute('id');
  if (id) {
    var link = document.createElement('a');
    link.classList.add('header-link');
    link.setAttribute('href', '#' + id);
    heading.appendChild(link);
  }
}
</script>

{--no-toc!}.+skip
<script>
function appendTocEntry(heading) {
  var id = heading.getAttribute('id');
  if (heading.classList.contains('no-auto-toc')) {
    return;
  }
  var container = document.getElementById('auto-toc');
  if (container === null) {
    return;
  }
  var tocLink = document.createElement('a');
  tocLink.setAttribute('href', '#' + id);
  tocLink.textContent = heading.textContent;
  var tocEntry = document.createElement('div');
  tocEntry.setAttribute('class', heading.tagName.toLowerCase());
  tocEntry.appendChild(tocLink);
  container.appendChild(tocEntry);
}
</script>

{--dropdown-toc=}.+skip
<script>
// matches() polyfill for old browsers.
if (!Element.prototype.matches) {
  var p = Element.prototype;
  if (p.webkitMatchesSelector) // Chrome <34, SF<7.1, iOS<8
    p.matches = p.webkitMatchesSelector;
  if (p.msMatchesSelector) // IE9/10/11 & Edge
    p.matches = p.msMatchesSelector;
  if (p.mozMatchesSelector) // FF<34
    p.matches = p.mozMatchesSelector;
}
window.onclick = function(event) {
  var body = document.getElementsByTagName('body')[0];
  if (event.target.matches('#toc-button, #toc a')) {
    // Toggle TOC if TOC button or TOC link is clicked.
    body.classList.toggle('show-toc');
  }
  else if (!event.target.matches('#toc, #toc *')) {
    // Hide TOC if clicked outside TOC.
    body.classList.remove('show-toc');
  }
}
</script>

</body>
</html>''',
    'v8-header.rmu': r'''/*
  Used by rimuc `--layout v8` option.
  DEPRECATED: This layout is no longer maintained, for Rimu version 8 compatibility.
  Styled using Bootstrap.
  Syntax highlighting with Highlight.js
  Bootstrap and Highlight.js sourced from CDNs.
*/

// Set macro default values.
{--highlightjs?} = ''
{--mathjax?} = ''
{--section-numbers?} = ''
{--lang?} = ''
{--title?} = '&nbsp;'
{--custom-toc?} = ''
{--theme?} = 'default'
{--sidebar-toc?} = ''
{--dropdown-toc?} = ''

// DEPRECATED --toc: If --toc is non-blank make --sidebar-toc non-blank.
{--toc?} = ''
{--sidebar-toc} = '{--toc}{--sidebar-toc}'

<!DOCTYPE HTML>
<html lang="{--lang}">
<head>
{--!} Force IE into Standards mode.
  <meta http-equiv="X-UA-Compatible" content="IE=edge">
  <meta charset="UTF-8">
  <title>{--title}</title>
  <link rel="stylesheet" href="https://netdna.bootstrapcdn.com/twitter-bootstrap/2.3.2/css/bootstrap-combined.min.css">
  <link rel="stylesheet" href="https://yandex.st/highlightjs/7.3/styles/default.min.css">

<style>
  /* Bootstrap tweaks. */
  body {
    margin: 1em;
    max-width: 50em;
    font-family: Arial, Helvetica, sans-serif;
  }
  h1, h2, h3, h4, h5, h6 {
    color: #527bbd;
  }
  h1 { font-size: 2.2em; }
  h2 { font-size: 1.5em; }
  h3 { font-size: 1.2em; }
  h4 { font-size: 1.1em; }
  h5 { font-size: 1.0em; }
  h6 { font-size: 0.9em; }
  h2, h3, h4, h5, h6 {
    line-height: normal;
    margin-top: 1.0em;
    margin-bottom: 0.2em;
  }
  h2 {
    margin-top: 1.5em;
    border-bottom: 1px solid rgba(0, 0, 0, 0.15);
  }
  a, a:hover {
    color: #527bbd;
  }
  li, dd {
    margin-bottom: 0.25em;
  }
  p, pre, li, dt, dd, blockquote p {
    font-size: inherit;
    line-height: 1.45;
    margin-top: 0.5em;
    margin-bottom: 0.5em;
  }
  blockquote {
    border: none;
    border-left: 5px solid #eeeeee;
    font-family: Georgia, serif;
    font-style: italic;
    width: 80%;
    margin: 1.5em 0;
    color: #383838;
  }
  .cite {
    font-style: italic;
    color:#777;
    padding:5px 0;
  }
  .cite::before {
    content: "\2014 \2009";
  }
  pre {
    padding: 6px;
    line-height: normal;
  }
  .light-background {
    background-color: #f8f8f8;
    border: none;
  }
  code {
    color: inherit;
    font-size: inherit;
  }
  *:not(pre) > code {
    border: none;
    background-color: #f8f8f8;
  }
  .dl-horizontal > dt {
    text-align: left;
    margin-top: 1.0em;
  }
  .dl-horizontal > dd {
    margin-top: 1.0em;
  }
  /* highlight.js tweaks. */
  pre > code {
    background-color: inherit;
    padding: 0;
  }
  pre span {
    opacity: 1 !important;
  }
  /* Rimu styles. */
{--!} Apply verse class to Normal Paragraphs and Division blocks.
  .verse {
    margin: 1.5em 20px;
  }
  div.verse p, p.verse {
    font-family: Georgia, serif;
    white-space:pre;
    margin-top: 0.75em;
    margin-bottom: 0.75em;
  }
{--!} Apply sidebar class to Normal Paragraphs and Division blocks.
  .sidebar {
    border: 1px solid silver;
    border-radius:4px;
    background: #ffffee;
    padding: 10px;
    margin: 1.5em 20px;
  }
  div.sidebar *:first-child {
    margin-top: 0.2em;
  }
  /* List item counters and definition list numbering. */
{--!} List item counters.
  dl {
    counter-reset: dl-counter;
  }
  dl > dt {
    counter-increment: dl-counter;
  }
  ol {
    counter-reset: ol-counter;
  }
  ol > li {
    counter-increment: ol-counter;
  }
  ul {
    counter-reset: ul-counter;
  }
  ul > li {
    counter-increment: ul-counter;
  }
{--!} Prefix list counter to counter class element content.
  .dl-counter:before {
    content: counter(dl-counter) " ";
  }
  .ol-counter:before {
    content: counter(ol-counter) " ";
  }
  .ul-counter:before {
    content: counter(ul-counter) " ";
  }
{--!} Number labeled list items.
  .dl-numbered > dt:before {
    content: counter(dl-counter) ". ";
  }
{--!} Force page break before the element.
  .page-break {
    page-break-before: always;
  }
{--!} Avoid page breaks inside the element.
  .no-page-break {
    page-break-inside: avoid;
  }
{--!} Text block alignment classes.
  .align-left {
    text-align: left;
  }
  .align-center {
    text-align: center;
  }
  .align-right {
    text-align: right;
  }
{--!} Do not wrap line breaks.
  .line-breaks {
    white-space:pre;
  }
</style>

{--section-numbers=}.+skip
<style>
  /* Section numbers. */
  body,h1 { counter-reset: h2-counter; }
  h2      { counter-reset: h3-counter; }
  #contents > h2:before {
    content: counter(h2-counter) ". ";
    counter-increment: h2-counter;
  }
  #contents > h3:before {
    content: counter(h2-counter) "." counter(h3-counter) ". ";
    counter-increment: h3-counter;
  }
</style>

{--sidebar-toc=}.+skip
<style>
  body {
    padding-left: 21em;
  }
  #toc {
    position: fixed;
    top: 0;
    left: 0;
    bottom: 0;
    width: 18em;
    border-right: 1px solid #cccccc;
    overflow-y: auto;
    overflow-x: hidden;
    box-shadow: 0 0 3px rgba(0, 0, 0, 0.35);
  }
</style>

{--dropdown-toc=}.+skip
<style>
  .toc-visible {
    display: block !important;
  }
  #toc-button {
    position: fixed;
    top: 22px;
    left: 15px;
    z-index: 1;
    cursor: pointer;
    color: silver;
    font-size:3.2em;
  }
  #toc {
    display: none;
    position: fixed;
    top: 55px;
    left: 15px;
    right: 10%;
    z-index: 1;
    max-width: 30em;
    max-height: 80%;
    overflow-y: auto;
    background-color: white;
    box-shadow: 0 0 10px rgba(0, 0, 0, 0.35);
  }
  #contents {
    margin-left: 40px;
  }
</style>

// Common to --sidebar-toc and --dropdown-toc.
.+skip
{--sidebar-toc!}.-skip
{--dropdown-toc!}.-skip
<style>
  @media print {
    .no-print, .no-print * {
      display: none !important;
    }
    body {
      padding-left: 1em;
    }
  }
  #toc {
    padding-left: 1em;
  }
  #toc .h1 {
    font-size: 110%;
    font-weight: bold;
    margin-top: 0.5em;
    margin-bottom: 0.4em;
  }
  #toc .h2 {
    margin-top: 0.4em;
  }
  #toc .h3 {
    margin-left: 1.5em;
    font-size: 90%;
  }
  #toc div:nth-child(even) {
    background-color: #f8f8f8;
  }
</style>

{--theme!.*\bgraystone\b.*}.+skip
<style>
  body {
    font-size: 13pt;
  }
  h1, h2, h3, h4, h5, h6 {
    color: #888;
  }
  h1, h2 {
    text-transform: uppercase;
  }
  a, a:hover {
    color: #888;
    text-decoration: underline;
  }
  @media print {
    a {
      text-decoration: none;
    }
  }
</style>

</head>
<body>

// Include dropdown TOC button unless a custom TOC is specified.
{--dropdown-toc=}.+skip
{--custom-toc!}.+skip
<div id="toc-button" onclick="toggleToc()" class="no-print">&#8801;</div>

// Include for sidebar and dropdown TOC unless a custom TOC is specified.
.+skip
{--sidebar-toc!}.-skip
{--dropdown-toc!}.-skip
{--custom-toc!}.+skip
<div id="toc" class="no-print">
  <div id="auto-toc"></div>
</div>

<div id="contents">''',
    'flex-header.rmu': r'''/*
  Used by rimuc `--layout flex` option.
*/

/*
  Default values for styling themes and content options.
*/

{--section-numbers?} = ''
{--lang?} = ''
{--title?} = '&nbsp;'
{--theme?} = ''
{--no-toc?} = ''
{--custom-toc?} = ''
{--header-links?} = ''
{--!} The min-width: 1px clause stops page load transitions in IE11 and Edge (https://stackoverflow.com/a/25850649).
{--small-screen?} = '(min-width: 1px) and (max-width: 800px)'
{--meta?} = '<meta charset="UTF-8">
  {--!} Make old IE versions use the latest rendering engine.
  <meta http-equiv="X-UA-Compatible" content="IE=edge">
  <meta name="viewport" content="width=device-width, initial-scale=1, minimum-scale=1">'
// Additional <head> child elements.
{--head?} = ''

{--highlightjs?} = ''
{--highlightjs-css} = '<link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.5.0/styles/default.min.css">'
{--highlightjs-scripts} = '<script src="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.5.0/highlight.min.js"></script>
<script>
hljs.configure({languages: []});
hljs.highlightAll();
</script>'

{--mathjax?} = ''
{--mathjax-scripts} = '<script async src="https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.2/MathJax.js?config=TeX-MML-AM_CHTML"></script>'

// List item CSS counters.
{--dl-counter} = '<span class="dl-counter"></span>'
{--ol-counter} = '<span class="ol-counter"></span>'
{--ul-counter} = '<span class="ul-counter"></span>'

/*
  Legend theme.
*/
{--theme!.*\blegend\b.*}.+skip
..
{--sans-font?} = 'Helvetica, Arial, sans-serif'
{--serif-font?} = 'Georgia, Times, serif'
{--mono-font?} = 'Consolas, "Andale Mono", Monaco, monospace'
{--mono-size?} = '90%'
{--text-color?} = '#333333'
{--border-color?} = '#dddddd'
{--primary-color?} = '#34495e'
{--primary-background?} = 'white'
{--code-background?} = '#f8f8f8'
{--sidebar-background?} = '#ffffee'
{--link-color?} = '#428bca'
{--border-radius?} = '0'
{--max-text-width?} = '50rem'
..

/*
  Vintage theme.
*/
{--theme!.*\bvintage\b.*}.+skip
// Default to this theme if none of the built-in themes is specified.
{--theme!.*\b(legend|vintage|graystone)\b.*}.-skip
..
{--sans-font?} = 'Helvetica, Arial, sans-serif'
{--serif-font?} = 'Georgia, Times, serif'
{--mono-font?} = 'Consolas, "Andale Mono", Monaco, monospace'
{--mono-size?} = '90%'
{--text-color?} = '#333333'
{--border-color?} = '#dddddd'
{--primary-color?} = '#527bbd'
{--primary-background?} = 'white'
{--code-background?} = '#f8f8f8'
{--sidebar-background?} = '#ffffee'
{--link-color?} = '#527bbd'
{--border-radius?} = '4px'
{--max-text-width?} = '50rem'
..

/*
  Graystone theme.
*/
{--theme!.*\bgraystone\b.*}.+skip
..
{--sans-font?} = 'Helvetica, Arial, sans-serif'
{--serif-font?} = 'Georgia, Times, serif'
{--mono-font?} = 'Consolas, "Andale Mono", Monaco, monospace'
{--mono-size?} = '90%'
{--text-color?} = '#333333'
{--border-color?} = '#dddddd'
{--primary-color?} = '#888'
{--primary-background?} = 'white'
{--code-background?} = '#f8f8f8'
{--sidebar-background?} = '#ffffee'
{--link-color?} = '#888'
{--border-radius?} = '0'
{--max-text-width?} = '50rem'
..

/*
  HTML header
*/
<!DOCTYPE HTML>
{--lang=}<html>
{--lang!}<html lang="{--lang}">
<head>
{--meta}
<title>{--title}</title>
{--highlightjs!}{--highlightjs-css}

/*
  Layout independent styles
*/
<style>
  :root {
    font-size: 16px;
    line-height: 20px;
    color: {--text-color};
    background-color: {--primary-background};
    font-family: {--sans-font};
  }
  * {
    margin: 0;
    padding: 0;
    white-space: inherit;
    box-sizing: border-box;
  }
  h1, h2, h3, h4, h5, h6 {
    margin: 10px 0;
    color: {--primary-color};
  }
  h1 { font-size: 2.1rem; line-height: 40px;}
  h2 { font-size: 1.5rem; }
  h3 { font-size: 1.4rem; }
  h4 { font-size: 1.1rem; }
  h5 { font-size: 1.0rem; }
  h6 { font-size: 0.9rem; }
  h2, h3, h4, h5, h6 {
    line-height: normal;
    margin-top: 1.0rem;
    margin-bottom: 0.2rem;
  }
  h2 {
    margin-top: 1.5rem;
    border-bottom: 1px solid {--border-color};
  }
  img {
    max-width: 100%;
  }
  table {
    border-collapse: collapse;
    table-layout: fixed;
    width: 100%;
  }
  td, th {
    text-align: left;
    vertical-align: top;
  }
  {--!} tbody necessary because: https://stackoverflow.com/questions/7490364/why-do-browsers-still-inject-tbody-in-html5
  table.bordered > tbody > tr > th,
  table.bordered > tbody > tr > td {
    border: 1px solid {--border-color};
  }
  a, a:hover {
    color: {--primary-color};
  }
  a {
    text-decoration: none;
  }
  a:hover {
    text-decoration: underline;
  }
  p, pre, li, dt, dd, blockquote p {
    line-height: 1.45;
    margin-top: 0.5rem;
    margin-bottom: 0.5rem;
  }
  ul, ol {
    padding: 0;
    margin: 0 0 10px 25px;
  }
  dt {
    margin-bottom: 0;
    font-style: italic;
  }
  dd {
    margin-left: 1.0rem;
    margin-top: 0;
  }
  blockquote {
    padding: 0 0 0 15px;
    border: none;
    border-left: 5px solid #eeeeee;
    font-family: {--serif-font};
    width: 80%;
    margin: 1.5em 0;
    color: #383838;
  }
  .cite {
    color: #777;
    padding: 5px 0;
  }
  .cite::before {
    content: "\2014 \2009";
  }
  code {
    font-family: {--mono-font};
    {--!} Monspace fonts are relatively oversized.
    font-size: {--mono-size};
    background-color: {--code-background};
  }
  pre {
    font-family: {--mono-font};
    line-height: normal;
    white-space: pre-wrap;
    background-color: {--code-background};
    border: 1px solid {--border-color};
    border-radius: {--border-radius};
    padding: 0;
  }
  .light-background {
    background-color: {--code-background};
    border: none;
    white-space: pre-wrap;
  }
  .dl-horizontal > dd {
    margin-top: 1.0rem;
  }
  pre > code {
    background-color: inherit;
    display: block;
    padding: 0.5em !important;
    border-radius: {--border-radius};
  }
  pre span {
    {--!} highlight.js tweak.
    opacity: 1 !important;
  }
  {--!} Rimu classes.
  {--!} Apply verse class to Normal Paragraphs and Division blocks.
  .verse {
    margin: 1.5em 20px;
  }
  div.verse p, p.verse {
    font-family: {--serif-font};
    white-space: pre-wrap;
    margin-top: 0.75rem;
    margin-bottom: 0.75rem;
  }
  {--!} Apply sidebar and admonition classes to Normal Paragraphs and Division blocks.
  .sidebar,
  .important,
  .note,
  .tip,
  .warning {
    margin: 1.5em 0;
    padding: 10px;
    border-radius: {--border-radius};
  }
  div.sidebar *:first-child,
  div.important *:first-child,
  div.note *:first-child,
  div.tip *:first-child,
  div.warning *:first-child {
    margin-top: 0.2rem !important;
  }
  .note::before,
  .important::before,
  .tip::before,
  .warning::before {
    font-weight: bold;
  }
  .notitle::before {
    content: "" !important;
  }
  .sidebar {
    background: #ffffee;
    border: 1px solid {--border-color};
  }
  .note {
    background-color: #f0f7fb;
    border-left: solid 4px #3498db;
  }
  div.note::before {
    content: "Note";
  }
  p.note::before {
    content: "Note: ";
  }
  .tip {
    background-color: #e7f6ef;
    border-left: solid 4px #32c875;
  }
  div.tip::before {
    content: "Tip";
  }
  p.tip::before {
    content: "Tip: ";
  }
  .warning {
    background-color: #fdf7f2;
    border-left: solid 4px #d1534a;
  }
  div.warning::before {
    content: "Warning";
  }
  p.warning::before {
    content: "Warning: ";
  }
  .important {
    background-color: #fffbea;
    border-left: solid 4px #eec51c;
  }
  div.important::before {
    content: "Important";
  }
  p.important::before {
    content: "Important: ";
  }
  {--!} Force page break before the element.
  .page-break {
    page-break-before: always;
  }
  {--!} Avoid page breaks inside the element.
  .no-page-break {
    page-break-inside: avoid;
  }
  {--!} Text block alignment classes.
  .align-left {
    text-align: left;
  }
  .align-center {
    text-align: center;
  }
  .align-right {
    text-align: right;
  }
  {--!} Preserve line breaks.
  .preserve-breaks {
    white-space: pre;
  }
  {--!} DEPRECATED: Use `preserve-breaks` instead.
  .line-breaks {
    white-space: pre-wrap;
  }
  {--!} Horizontal labeled list.
  .dl-horizontal:before, .dl-horizontal:after {
    display: table;
    content: "";
    line-height: 0;
  }
  .dl-horizontal:after {
    clear: both;
  }
  .dl-horizontal > dt {
    float: left;
    clear: left;
    overflow: hidden;
    text-overflow: ellipsis;
    white-space: nowrap;
    text-align: left;
    margin-top: 1.0rem;
    width: 25%;
  }
  .dl-horizontal > dd {
    margin-left: 25%;
    padding-left: 1.0rem;
  }
  {--!} List item counters.
  dl {
    counter-reset: dl-counter;
  }
  dl > dt {
    counter-increment: dl-counter;
  }
  ol {
    counter-reset: ol-counter;
  }
  ol > li {
    counter-increment: ol-counter;
  }
  ul {
    counter-reset: ul-counter;
  }
  ul > li {
    counter-increment: ul-counter;
  }
  {--!} Prefix list counter to counter class element content.
  .dl-counter:before {
    content: counter(dl-counter) " ";
  }
  .ol-counter:before {
    content: counter(ol-counter) " ";
  }
  .ul-counter:before {
    content: counter(ul-counter) " ";
  }
  {--!} Number labeled list items.
  .dl-numbered > dt:before {
    content: counter(dl-counter) ". ";
  }
  {--!} Printing.
  @media print {
    .no-print, .no-print * {
      display: none !important;
    }
    a:hover {
      text-decoration: none !important;
    }
  }
</style>

{--section-numbers=}.+skip
<style>
  /* Section numbers. */
  body,h1 { counter-reset: h2-counter; }
  h2      { counter-reset: h3-counter; }
  #article > h2:before {
    content: counter(h2-counter) ". ";
    counter-increment: h2-counter;
  }
  #article > h3:before {
    content: counter(h2-counter) "." counter(h3-counter) ". ";
    counter-increment: h3-counter;
  }
</style>

{--no-toc!}.+skip
<style>
  #toc .h1 {
    font-size: 110%;
    font-weight: bold;
    margin-top: 0.5rem;
    margin-bottom: 0.4rem;
  }
  #toc .h2 {
    margin-top: 0.4rem;
  }
  #toc .h3 {
    margin-left: 1.5rem;
    font-size: 90%;
  }
  #toc div[class^="h"]:nth-child(even) {
    background-color: #f8f8f8;
  }
  #auto-toc {
    width: 100%;
  }
</style>

{--header-links=}.+skip
<style>
  .header-link {
    text-decoration: none;
    color: lightgray;
    visibility: hidden;
  }
  .header-link:hover {
    text-decoration: none;
    color: gray;
  }
  .header-link::before {
    content: " \00b6";
  }
  h2:hover > .header-link, h3:hover > .header-link {
    visibility: visible;
  }
  @media print {
    .header-link {
      display: none !important;
    }
  }
</style>

/*
  Layout styles
*/
<style>
  #article > * {
    max-width: {--max-text-width};
  }
  #article, #toc {
    padding-left: 1rem;
    padding-right: 1rem;
  }
{--!} Small screen.
  @media screen and {--small-screen} {
    #article, #toc {
      padding-left: 10px;
      padding-right: 10px;
    }
  }
</style>

// TOC specific styles.
// Flexbox layout inspired by https://codepen.io/VinSpee/pen/zxBJVO
{--no-toc!}.+skip
<style>
  html, body {
    height: 100%;
    min-height: 100%;
    overflow: hidden;
  }
  body {
    display: flex;
  }
  #nav {
    flex: auto;
    order: 1;
    flex-grow: 0;
    background: {--primary-color};
  }
  #toc {
    display: none;
    flex: auto;
    order: 2;
    overflow-y: scroll;
    height: 100%;
    width: 100%;
  }
  #toc h2 {
    font-size: 125%;
    margin-top:1.0rem;
  }
  #toc-button {
    cursor: pointer;
    color: white;
    font-size: 48px;
    font-family: Arial;
    display: inline-block;
  }
  #article {
    flex: auto;
    order: 3;
    overflow-y: scroll;
    width: 100%;
  }
  #close-icon {
    display: none;
  }
  .show-toc #toc {
    display: block !important;
  }
  .show-toc #article {
    display: none !important;
  }
  .show-toc #menu-icon {
    display: none !important;
  }
  .show-toc #close-icon {
    display: inline !important;
  }
  @media print {
    html, body {
      height: auto;
      min-height: 0;
    }
    #article, #toc {
      overflow-y: visible;
    }
  }
</style>

{--no-toc!}.+skip
<style>
{--!} Side-bar for large screen and small screen landscape mode.
  body {
    flex-direction: row;
  }
  #nav {
    height: 100%;
    width: 40px;
  }
  #toc-button {
    padding-left: 4px;
    padding-top: 15px;
  }
{--!} Top-bar for small screen portrait mode.
  @media screen and {--small-screen} and (orientation: portrait) {
    body {
      flex-direction: column;
    }
    #nav {
      height: 40px;
      width: 100%;
    }
    #toc-button {
      padding-left: 10px;
      padding-top: 10px;
    }
  }
</style>

// Format (non-layout) related styles.
<style>
  @media screen and {--small-screen} {
    body {
      font-size: 20px;
    }
    * {
      overflow-wrap: break-word;
    }
  }
</style>

/*
  Theme related styles
*/
{--theme!.*\bgraystone\b.*}.+skip
<style>
  h1, h2 {
    text-transform: uppercase;
  }
</style>

{--head}

</head>
<body>

// Include TOC menu bar if a TOC is specified.
{--no-toc!}.+skip
<div id="nav" class="no-print">
  <span id="toc-button">
    <span id="menu-icon">&#8801;</span>
    <span id="close-icon">&times;</span>
  </span>
</div>

// Include TOC unless a custom TOC is specified.
.+skip
{--no-toc=}.-skip
{--custom-toc!}.+skip
<div id="toc">
  <div id="auto-toc"></div>
</div>

<div id="article">''',
    'sequel-header.rmu': r'''/*
  Used by rimuc `--layout sequel` option.
*/

/*
  Default values for styling themes and content options.
*/

{--section-numbers?} = ''
{--lang?} = ''
{--title?} = '&nbsp;'
{--theme?} = ''
{--no-toc?} = ''
{--custom-toc?} = ''
{--header-links?} = ''
{--!} The min-width: 1px clause stops page load transitions in IE11 and Edge (https://stackoverflow.com/a/25850649).
{--small-screen?} = '(min-width: 1px) and (max-width: 800px)'
{--meta?} = '<meta charset="UTF-8">
  {--!} Make old IE versions use the latest rendering engine.
  <meta http-equiv="X-UA-Compatible" content="IE=edge">
  <meta name="viewport" content="width=device-width, initial-scale=1, minimum-scale=1">'
// Additional <head> element children.
{--head?} = ''

{--highlightjs?} = ''
{--highlightjs-css} = '<link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.5.0/styles/default.min.css">'
{--highlightjs-scripts} = '<script src="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.5.0/highlight.min.js"></script>
<script>
hljs.configure({languages: []});
hljs.highlightAll();
</script>'

{--mathjax?} = ''
{--mathjax-scripts} = '<script async src="https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.2/MathJax.js?config=TeX-MML-AM_CHTML"></script>'

// List item CSS counters.
{--dl-counter} = '<span class="dl-counter"></span>'
{--ol-counter} = '<span class="ol-counter"></span>'
{--ul-counter} = '<span class="ul-counter"></span>'

// Sequel layout specific.
{--toc-width?} = '300px'
{--nav-width?} = '40px'
{--content-margin?} = '340px'
{--transition-duration?} = '0.2s'

/*
  Legend theme.
*/
{--theme!.*\blegend\b.*}.+skip
// Default to this theme if none of the built-in themes is specified.
{--theme!.*\b(legend|vintage|graystone)\b.*}.-skip
..
{--sans-font?} = 'Helvetica, Arial, sans-serif'
{--serif-font?} = 'Georgia, Times, serif'
{--mono-font?} = 'Consolas, "Andale Mono", Monaco, monospace'
{--mono-size?} = '90%'
{--text-color?} = '#333333'
{--border-color?} = '#dddddd'
{--primary-color?} = '#34495e'
{--primary-background?} = 'white'
{--code-background?} = '#f8f8f8'
{--sidebar-background?} = '#ffffee'
{--link-color?} = '#428bca'
{--border-radius?} = '0'
{--max-text-width?} = '55rem'
..

/*
  Vintage theme.
*/
{--theme!.*\bvintage\b.*}.+skip
..
{--sans-font?} = 'Helvetica, Arial, sans-serif'
{--serif-font?} = 'Georgia, Times, serif'
{--mono-font?} = 'Consolas, "Andale Mono", Monaco, monospace'
{--mono-size?} = '90%'
{--text-color?} = '#333333'
{--border-color?} = '#dddddd'
{--primary-color?} = '#527bbd'
{--primary-background?} = 'white'
{--code-background?} = '#f8f8f8'
{--sidebar-background?} = '#ffffee'
{--link-color?} = '#527bbd'
{--border-radius?} = '4px'
{--max-text-width?} = '55rem'
..

/*
  Graystone theme.
*/
{--theme!.*\bgraystone\b.*}.+skip
..
{--sans-font?} = 'Helvetica, Arial, sans-serif'
{--serif-font?} = 'Georgia, Times, serif'
{--mono-font?} = 'Consolas, "Andale Mono", Monaco, monospace'
{--mono-size?} = '90%'
{--text-color?} = '#333333'
{--border-color?} = '#dddddd'
{--primary-color?} = '#888'
{--primary-background?} = 'white'
{--code-background?} = '#f8f8f8'
{--sidebar-background?} = '#ffffee'
{--link-color?} = '#888'
{--border-radius?} = '0'
{--max-text-width?} = '55rem'
..

/*
  HTML header
*/
<!DOCTYPE HTML>
{--lang=}<html>
{--lang!}<html lang="{--lang}">
<head>
{--meta}
<title>{--title}</title>
{--highlightjs!}{--highlightjs-css}

/*
  Layout independent styles
*/
<style>
  :root {
    font-size: 16px;
    line-height: 20px;
    color: {--text-color};
    background-color: {--primary-background};
    font-family: {--sans-font};
  }
  * {
    margin: 0;
    padding: 0;
    white-space: inherit;
    box-sizing: border-box;
  }
  h1, h2, h3, h4, h5, h6 {
    margin: 10px 0;
    color: {--primary-color};
  }
  h1 { font-size: 2.1rem; line-height: 40px;}
  h2 { font-size: 1.5rem; }
  h3 { font-size: 1.4rem; }
  h4 { font-size: 1.1rem; }
  h5 { font-size: 1.0rem; }
  h6 { font-size: 0.9rem; }
  h2, h3, h4, h5, h6 {
    line-height: normal;
    margin-top: 1.0rem;
    margin-bottom: 0.2rem;
  }
  h2 {
    margin-top: 1.5rem;
    border-bottom: 1px solid {--border-color};
  }
  img {
    max-width: 100%;
  }
  table {
    border-collapse: collapse;
    table-layout: fixed;
    width: 100%;
  }
  td, th {
    text-align: left;
    vertical-align: top;
  }
  {--!} tbody necessary because: https://stackoverflow.com/questions/7490364/why-do-browsers-still-inject-tbody-in-html5
  table.bordered > tbody > tr > th,
  table.bordered > tbody > tr > td {
    border: 1px solid {--border-color};
  }
  a, a:hover {
    color: {--link-color};
  }
  a {
    text-decoration: none;
  }
  a:hover {
    text-decoration: underline;
  }
  p, pre, li, dt, dd, blockquote p {
    line-height: 1.45;
    margin-top: 0.5rem;
    margin-bottom: 0.5rem;
  }
  ul, ol {
    padding: 0;
    margin: 0 0 10px 25px;
  }
  dt {
    margin-bottom: 0;
    font-style: italic;
  }
  dd {
    margin-left: 1.0rem;
    margin-top: 0;
  }
  blockquote {
    padding: 0 0 0 15px;
    border: none;
    border-left: 5px solid #eeeeee;
    font-family: {--serif-font};
    width: 80%;
    margin: 1.5em 0;
    color: #383838;
  }
  .cite {
    color: #777;
    padding: 5px 0;
  }
  .cite::before {
    content: "\2014 \2009";
  }
  code {
    font-family: {--mono-font};
    {--!} Monspace fonts are relatively oversized.
    font-size: {--mono-size};
    background-color: {--code-background};
  }
  pre {
    font-family: {--mono-font};
    line-height: normal;
    white-space: pre-wrap;
    background-color: {--code-background};
    border: 1px solid {--border-color};
    border-radius: {--border-radius};
    padding: 0;
  }
  .light-background {
    background-color: {--code-background};
    border: none;
    white-space: pre-wrap;
  }
{--!} Do not break hyphenated words or white space inside code quotes.
  *:not(pre) > code {
    white-space: nowrap;
  }
  .dl-horizontal > dd {
    margin-top: 1.0rem;
  }
  pre > code {
    background-color: inherit;
    display: block;
    padding: 0.5em !important;
    border-radius: {--border-radius};
  }
  pre span {
    {--!} highlight.js tweak.
    opacity: 1 !important;
  }
  {--!} Rimu classes.
  {--!} Apply verse class to Normal Paragraphs and Division blocks.
  .verse {
    margin: 1.5em 20px;
  }
  div.verse p, p.verse {
    font-family: {--serif-font};
    white-space: pre-wrap;
    margin-top: 0.75rem;
    margin-bottom: 0.75rem;
  }
  {--!} Apply sidebar and admonition classes to Normal Paragraphs and Division blocks.
  .sidebar,
  .important,
  .note,
  .tip,
  .warning {
    margin: 1.5em 0;
    padding: 10px;
    border-radius: {--border-radius};
  }
  div.sidebar *:first-child,
  div.important *:first-child,
  div.note *:first-child,
  div.tip *:first-child,
  div.warning *:first-child {
    margin-top: 0.2rem !important;
  }
  .note::before,
  .important::before,
  .tip::before,
  .warning::before {
    font-weight: bold;
  }
  .notitle::before {
    content: "" !important;
  }
  .sidebar {
    background: #ffffee;
    border: 1px solid {--border-color};
  }
  .note {
    background-color: #f0f7fb;
    border-left: solid 4px #3498db;
  }
  div.note::before {
    content: "Note";
  }
  p.note::before {
    content: "Note: ";
  }
  .tip {
    background-color: #e7f6ef;
    border-left: solid 4px #32c875;
  }
  div.tip::before {
    content: "Tip";
  }
  p.tip::before {
    content: "Tip: ";
  }
  .warning {
    background-color: #fdf7f2;
    border-left: solid 4px #d1534a;
  }
  div.warning::before {
    content: "Warning";
  }
  p.warning::before {
    content: "Warning: ";
  }
  .important {
    background-color: #fffbea;
    border-left: solid 4px #eec51c;
  }
  div.important::before {
    content: "Important";
  }
  p.important::before {
    content: "Important: ";
  }
  {--!} Force page break before the element.
  .page-break {
    page-break-before: always;
  }
  {--!} Avoid page breaks inside the element.
  .no-page-break {
    page-break-inside: avoid;
  }
  {--!} Text block alignment classes.
  .align-left {
    text-align: left;
  }
  .align-center {
    text-align: center;
  }
  .align-right {
    text-align: right;
  }
  {--!} Preserve line breaks.
  .preserve-breaks {
    white-space: pre;
  }
  {--!} DEPRECATED: Use `preserve-breaks` instead.
  .line-breaks {
    white-space: pre-wrap;
  }
  {--!} Horizontal labeled list.
  .dl-horizontal:before, .dl-horizontal:after {
    display: table;
    content: "";
    line-height: 0;
  }
  .dl-horizontal:after {
    clear: both;
  }
  .dl-horizontal > dt {
    float: left;
    clear: left;
    overflow: hidden;
    text-overflow: ellipsis;
    white-space: nowrap;
    text-align: left;
    margin-top: 1.0rem;
    width: 25%;
  }
  .dl-horizontal > dd {
    margin-left: 25%;
    padding-left: 1.0rem;
  }
  {--!} List item counters.
  dl {
    counter-reset: dl-counter;
  }
  dl > dt {
    counter-increment: dl-counter;
  }
  ol {
    counter-reset: ol-counter;
  }
  ol > li {
    counter-increment: ol-counter;
  }
  ul {
    counter-reset: ul-counter;
  }
  ul > li {
    counter-increment: ul-counter;
  }
  {--!} Prefix list counter to counter class element content.
  .dl-counter:before {
    content: counter(dl-counter) " ";
  }
  .ol-counter:before {
    content: counter(ol-counter) " ";
  }
  .ul-counter:before {
    content: counter(ul-counter) " ";
  }
  {--!} Number labeled list items.
  .dl-numbered > dt:before {
    content: counter(dl-counter) ". ";
  }
  {--!} Printing.
  @media print {
    .no-print, .no-print * {
      display: none !important;
    }
    a:hover {
      text-decoration: none !important;
    }
  }
</style>

{--section-numbers=}.+skip
<style>
  /* Section numbers. */
  body,h1 { counter-reset: h2-counter; }
  h2      { counter-reset: h3-counter; }
  #article > h2:before {
    content: counter(h2-counter) ". ";
    counter-increment: h2-counter;
  }
  #article > h3:before {
    content: counter(h2-counter) "." counter(h3-counter) ". ";
    counter-increment: h3-counter;
  }
</style>

// TOC styling.
{--no-toc!}.+skip
<style>
  #toc h2 {
    font-size: 125%;
    padding-left: 1rem;
    margin-top:1.0rem;
    margin-left: 0;
  }
  #toc .h1 {
    font-size: 110%;
    font-weight: bold;
    padding-left: 1rem;
    margin-top: 0.5rem;
    margin-bottom: 0.4rem;
  }
  #toc .h2 {
    padding-left: 1rem;
    margin-top: 0.4rem;
  }
  #toc .h3 {
    padding-left: 2.5rem;
    font-size: 90%;
  }
  #toc div[class^="h"]:nth-child(even) {
    background-color: #f8f8f8;
  }
  #auto-toc {
    width: 100%;
  }
  #toc a, #toc a:hover {
    color: {--primary-color};
  }
</style>

{--header-links=}.+skip
<style>
  .header-link {
    text-decoration: none;
    color: lightgray;
    visibility: hidden;
  }
  .header-link:hover {
    text-decoration: none;
    color: gray;
  }
  .header-link::before {
    content: " \00b6";
  }
  h2:hover > .header-link, h3:hover > .header-link {
    visibility: visible;
  }
  @media print {
    .header-link {
      display: none !important;
    }
  }
</style>


/*
  Layout styles
*/
<style>
  #main {
    z-index: 1;
    min-height: 100%;
    max-width: {--max-text-width};
    margin-left: 0;
    padding-left: 40px;
    padding-right: 40px;
  }
  #article {
    padding-top: 1px;
  }
</style>

{--no-toc!}.+skip
<style>
  #main {
    margin-left: {--content-margin};
  }
  #nav {
    z-index: 3;
    position: fixed;
    left: 0;
    height: 100%;
    width: {--nav-width};
    background-color: {--primary-color};
    color: white;
  }
  #sidebar {
    z-index: 2;
    position: fixed;
    top: 0;
    left: 0;
    bottom: 0;
    margin-left: {--nav-width};
    width: {--toc-width};
    overflow: auto;
    border-right: 1px solid #ccc;
    background: #eee;
  }
  body.hide-toc #sidebar {
    width: 0;
  }
  body.hide-toc #main {
    margin-left: {--nav-width};
  }
  #toc-button {
    cursor: pointer;
    color: white;
    font-size: 48px;
    line-height: 40px;
    font-family: Arial, sans-serif;
    display: inline-block;
    padding-left: 5px;
  }
{--!} Apply TOC slide in/out transition when not in small screen mode.
  @media not screen and {--small-screen} {
    #sidebar {
      opacity: 1;
      transition: opacity {--transition-duration} ease;
    }
    body.hide-toc #sidebar {
      width: 0;
      opacity: 0;
    }
    #main {
      transition: margin-left {--transition-duration} ease;
    }
    body.hide-toc #main {
      transition: margin-left {--transition-duration} ease;
    }
  }
  @media screen and {--small-screen} {
    body:not(.hide-toc) {
{--!} Hide body Y scroll bar when #main content is hidden.
      overflow: hidden;
    }
    body:not(.hide-toc) #sidebar {
      width: 100%;
    }
    #main {
      margin-left: {--nav-width};
    }
    body:not(.hide-toc) #main {
      visibility: hidden;
    }
    #toc {
      margin-right: 40px;
    }
  }
</style>

<style>
  @media screen and {--small-screen} {
    body {
      font-size: 20px;
    }
    #main {
      padding-left: 10px;
      padding-right: 10px;
    }
    * {
      overflow-wrap: break-word;
    }
  }
</style>

<style>
  @media print {
    #main {
      margin-left: 0;
    }
  }
</style>

/*
  Theme related styles
*/
{--theme!.*\bgraystone\b.*}.+skip
<style>
  h1, h2 {
    text-transform: uppercase;
  }
</style>

{--head}

</head>
<body>

{--no-toc!}.+skip
<div id="nav" class="no-print">
    <span id="toc-button">&#8801;</span>
</div>

{--no-toc!}.+skip
<div id="sidebar" class="no-print">

// Include TOC unless no TOC or custom TOC is specified.
{--no-toc!}.+skip
{--custom-toc!}.+skip
<div id="toc" class="no-print">
  <div id="auto-toc"></div>
</div>

{--no-toc!}.+skip
</div>

<div id="main">
<div id="article">''',
    'plain-header.rmu': r'''/*
  Used by rimuc `--layout plain` option.
*/

{--lang?} = ''
{--title?} = 'Title'
{--meta?} = '<meta charset="UTF-8">'
// Additional <head> element children.
{--head?} = ''

<!DOCTYPE html>
{--lang=}<html>
{--lang!}<html lang="{--lang}">
<head>
{--meta}
<title>{--title}</title>

{--head}
</head>
<body>''',
    'flex-footer.rmu': r'''/*
  Used by rimuc.js --styled option.
*/

// Close article div.
</div>

{--highlightjs=}.+skip
{--highlightjs-scripts}

{--mathjax!}{--mathjax-scripts}

.+skip
{--no-toc=}.-skip
{--header-links!}.-skip
<script>
window.onload = function() {
{--no-toc=}  document.getElementsByTagName('body')[0].appendChild(document.getElementById('toc')); // Ensure custom TOC is child of body.
  var headings = [].slice.call(document.body.querySelectorAll('#article > h1, #article > h2, #article > h3'));
  headings.forEach(function(heading) {
{--header-links!}    setHeaderLink(heading);
{--no-toc=}    appendTocEntry(heading);
  });
}
</script>

{--header-links=}.+skip
<script>
function setHeaderLink(heading) {
  var id = heading.getAttribute('id');
  if (id) {
    var link = document.createElement('a');
    link.classList.add('header-link');
    link.setAttribute('href', '#' + id);
    heading.appendChild(link);
  }
}
</script>

{--no-toc!}.+skip
<script>
function appendTocEntry(heading) {
  var id = heading.getAttribute('id');
  if (heading.classList.contains('no-auto-toc')) {
    return;
  }
  var container = document.getElementById('auto-toc');
  if (container === null) {
    return;
  }
  var tocLink = document.createElement('a');
  tocLink.setAttribute('href', '#' + id);
  tocLink.textContent = heading.textContent;
  var tocEntry = document.createElement('div');
  tocEntry.setAttribute('class', heading.tagName.toLowerCase());
  tocEntry.appendChild(tocLink);
  container.appendChild(tocEntry);
}
// matches() polyfill for old browsers.
if (!Element.prototype.matches) {
  var p = Element.prototype;
  if (p.webkitMatchesSelector) // Chrome <34, SF<7.1, iOS<8
    p.matches = p.webkitMatchesSelector;
  if (p.msMatchesSelector) // IE9/10/11 & Edge
    p.matches = p.msMatchesSelector;
  if (p.mozMatchesSelector) // FF<34
    p.matches = p.mozMatchesSelector;
}
document.onclick = function(event) {
  if (event.target.matches('#toc-button *, #toc a')) {
    // Toggle TOC if TOC button or TOC link is clicked.
    document.getElementsByTagName('body')[0].classList.toggle('show-toc');
  }
}
</script>

</body>
</html>''',
}



from rimu import api, options
from rimu.options import RenderOptions


def render(source: str, opts: RenderOptions = None) -> str:
    '''Exported render() API.'''
    if opts is None:
        opts = RenderOptions()
    # Implicit first-call API initialisation.
    if options.safeMode == -1:
        api.init()
    options.updateFrom(opts)
    return api.render(source)



import re
from typing import List, Match, Optional, Pattern


class Reader:
    '''Rimu line oriented string reader.'''
    lines: List[str]
    pos: int    # Line index of current line.

    def __init__(self, text: str):
        # Used internally by spans package.
        text = text.replace('\u0000', ' ')
        # Used internally by spans package.
        text = text.replace('\u0001', ' ')
        # Used internally by macros package.
        text = text.replace('\u0002', ' ')
        # Split self.lines on newline boundaries.
        # http:#stackoverflow.com/questions/1155678/javascript-string-newline-character
        # Split is broken on IE8 e.g. 'X\n\nX'.split(/\n/g).length) returns 2 but should return 3.
        self.lines = re.split(r'\r\n|\r|\n', text)
        self.pos = 0

    @property
    def cursor(self) -> str:
        assert not self.eof()
        return self.lines[self.pos]

    @cursor.setter
    def cursor(self, value: str) -> None:
        assert not self.eof()
        self.lines[self.pos] = value

    def eof(self) -> bool:
        '''Return true if the cursor has advanced over all input self.lines.'''
        return self.pos >= len(self.lines)

    def next(self) -> None:
        '''Move cursor to next input line.'''
        if (not self.eof()):
            self.pos += 1

    def readTo(self, regexp: Pattern[str]) -> Optional[List[str]]:
        '''Read to the first line matching the regexp.

        Return the array of self.lines preceding the match plus a line containing
        the $1 match group (if it exists).
        Return null if an self.EOF is encountered.
        Exit with the reader pointing to the line following the match.'''
        result = []
        match = None
        while (not self.eof()):
            match = regexp.search(self.cursor)
            if (match is not None):
                if (regexp.groups > 0):
                    result.append(match[1])  # $1
                self.next()
                break
            result.append(self.cursor)
            self.next()
        # Blank line matches self.EOF.
        if (match is not None or (regexp.pattern == r'^$' and self.eof())):
            return result
        else:
            return None

    def skipBlankLines(self) -> None:
        while (not self.eof() and self.cursor.strip() == ''):
            self.next()


class Writer:
    '''Rimu line oriented string writer.'''
    buffer: List[str]

    def __init__(self):
        self.buffer = []

    def write(self, line: str) -> None:
        '''Write line.'''
        self.buffer.append(line)

    def toString(self) -> str:
        '''Return string of joined lines.'''
        return ''.join(self.buffer)



import re
from typing import Callable, List, Match, Optional, Pattern

from rimu import (blockattributes, delimitedblocks, expansion, io, macros,
                  options, quotes, replacements, utils)
from rimu.expansion import Expand

# Custom types.
Verify = Optional[Callable[[Match[str], io.Reader], bool]]   # Additional match verification checks.
Filter = Optional[Callable[[Match[str], io.Reader, 'Def'], str]]


class Def:
    match: Pattern[str]
    replacement: str
    name: str  # Optional unique identifier.
    verify: Verify
    filter: Filter

    def __init__(self,
                 match: Pattern[str],
                 replacement: str = '',
                 name: str = '',
                 verify: Verify = None,
                 filter: Filter = None,
                 ) -> None:
        self.match = match
        self.replacement = replacement
        self.name = name
        self.verify = verify
        self.filter = filter


def verifyMacroLine(match: Match[str], reader: io.Reader) -> bool:
    if macros.DEF_OPEN.search(match[0]):
        # Do not process macro definitions.
        return False
    # Silent because any macro expansion errors will be subsequently addressed downstream.
    value = macros.render(match[0], silent=True)
    if value.startswith(match[0]) or '\n' + match[0] in value:
        # The leading macro invocation expansion failed or contains itself.
        # This stops infinite recursion.
        return False
    # Insert the macro value into the reader just ahead of the cursor.
    pos = reader.pos+1
    reader.lines[pos:pos] = value.split('\n')
    return True


def blockDefFilter(match: Match[str], *_) -> str:
    if options.isSafeModeNz():
        return ''  # Skip if a safe mode is set.
    value = utils.replaceInline(match[2], Expand(macros=True))
    delimitedblocks.setDefinition(match[1], value)
    return ''


def quoteDefFilter(match: Match[str], *_) -> str:
    if options.isSafeModeNz():
        return ''  # Skip if a safe mode is set.
    quotes.setDefinition(quotes.Def(
        quote=match[1],
        openTag=utils.replaceInline(match[2], Expand(macros=True)),
        closeTag=utils.replaceInline(match[4], Expand(macros=True)),
        spans=match[3] == '|')
    )
    return ''


def replacementDefFilter(match: Match[str], *_) -> str:
    if options.isSafeModeNz():
        return ''  # Skip if a safe mode is set.
    pattern = match[1]
    flags = match[2]
    replacement = match[3]
    replacement = utils.replaceInline(replacement, Expand(macros=True))
    replacements.setDefinition(pattern, flags, replacement)
    return ''


def macroDefFilter(match: Match[str], *_) -> str:
    name = match[1]
    value = match[2]
    value = utils.replaceInline(value, Expand(macros=True))
    macros.setValue(name, value)
    return ''


def headerFilter(match: Match[str], _, d: Def) -> str:
    if macros.getValue('--header-ids') and blockattributes.id == '':
        blockattributes.id = blockattributes.slugify(match[2])
    result = utils.replaceMatch(match, d.replacement, Expand(macros=True))
    # Replace $1 with header number e.g. "<h###>" -> "<h3>"
    result = result.replace(match[1] + '>', str(len(match[1])) + '>')
    return result


def anchorFilter(match: Match[str], _, d: Def) -> str:
    if options.skipBlockAttributes():
        return ''
    else:
        # Default(non-filter) replacement processing.
        return utils.replaceMatch(match, d.replacement, Expand(macros=True))


def apiOptionFilter(match: Match[str], *_) -> str:
    if not options.isSafeModeNz():
        value = utils.replaceInline(match[2], Expand(macros=True))
        options.setOption(match[1], value)
    return ''


defs: List[Def] = [
    # Comment line.
    Def(
        match=re.compile(r'^\\?\/{2}(.*)$'),
    ),
    # Expand lines prefixed with a macro invocation prior to all other processing.
    # macro name = $1, macro value = $2
    Def(
        match=macros.MATCH_LINE,
        verify=verifyMacroLine,
        filter=lambda *_:'',     # Already processed in the `verify` function.
    ),
    # Delimited Block definition.
    # name = $1, definition = $2
    Def(
        match=re.compile(r"^\\?\|([\w\-]+)\|\s*=\s*'(.*)'$"),
        filter=blockDefFilter,
    ),
    # Quote definition.
    # quote = $1, openTag = $2, separator = $3, closeTag = $4
    Def(
        match=re.compile(r"^(\S{1,2})\s*=\s*'([^|]*)(\|{1,2})(.*)'$"),
        filter=quoteDefFilter,
    ),
    # Replacement definition.
    # pattern = $1, flags = $2, replacement = $3
    Def(
        match=re.compile(r"^\\?\/(.+)\/([igm]*)\s*=\s*'(.*)'$"),
        filter=replacementDefFilter,
    ),
    # Macro definition.
    # name = $1, value = $2
    Def(
        match=macros.LINE_DEF,
        filter=macroDefFilter,
    ),
    # Headers.
    # $1 is ID, $2 is header text.
    Def(
        match=re.compile(r'^\\?([#=]{1,6})\s+(.+?)(?:\s+\1)?$'),
        replacement=r'<h$1>$$2</h$1>',
        filter=headerFilter,
    ),
    # Block image: < image: src | alt >
    # src = $1, alt = $2
    Def(
        match=re.compile(r'^\\?<image:([^\s|]+)\|(.+?)>$'),
        replacement=r'<img src="$1" alt="$2">',
    ),
    # Block image: < image: src >
    # src = $1, alt = $1
    Def(
        match=re.compile(r'^\\?<image:([^\s|]+?)>$'),
        replacement=r'<img src="$1" alt="$1">',
    ),
    # DEPRECATED as of 3.4.0.
    # Block anchor: <<  # id>>
    # id = $1
    Def(
        match=re.compile(r'^\\?<<#([a-zA-Z][\w\-]*)>>$'),
        replacement=r'<div id="$1"></div>',
        filter=anchorFilter,
    ),
    # Block Attributes.
    # Syntax: .class-names  # id [html-attributes] block-options
    Def(
        name='attributes',
        # A loose match because Block Attributes can contain macro references.
        match=re.compile(r'^\\?\.[a-zA-Z#"\[+-].*$'),
        verify=lambda match, _: blockattributes.parse(match[0])
    ),
    # API Option.
    # name = $1, value = $2
    Def(
        match=re.compile(r"^\\?\.(\w+)\s*=\s*'(.*)'$"),
        filter=apiOptionFilter,
    ),
]


def render(reader: io.Reader, writer: io.Writer, allowed: List[str] = None) -> bool:
    '''If the next element in the reader is a valid line block render it
       and return True, else return false.'''
    if allowed is None:
        allowed = []
    if reader.eof():
        options.panic('premature eof')
    for d in defs:
        if allowed and d.name not in allowed:
            continue
        match = d.match.search(reader.cursor)
        if match is not None:
            if match[0][0] == '\\':
                # Drop backslash escape and continue.
                reader.cursor = reader.cursor[1:]
                continue
            if d.verify and not d.verify(match, reader):
                continue
            text: str
            if d.filter:
                text = d.filter(match, reader, d)
            else:
                text = utils.replaceMatch(match, d.replacement, Expand(macros=True)) if d.replacement else ''
            if text:
                text = blockattributes.injectHtmlAttributes(text)
                writer.write(text)
                reader.next()
                if not reader.eof():
                    writer.write('\n')  # Add a trailing '\n' if there are more lines.
            else:
                reader.next()
            return True
    return False



import re
from typing import List, Optional

from rimu import options, spans

# Matches a line starting with a macro invocation. $1 = macro invocation.
MATCH_LINE = re.compile(r'^({(?:[\w\-]+)(?:[!=|?](?:|.*?[^\\]))?}).*$')
# Match single-line macro definition. $1 = name, $2 = delimiter, $3 = value.
LINE_DEF = re.compile(r"^\\?{([\w\-]+\??)}\s*=\s*'(.*)'$")
# Match multi-line macro definition literal value open delimiter. $1 is first line of macro.
DEF_OPEN = re.compile(r"^\\?{[\w\-]+\??}\s*=\s*'(.*)$")
DEF_CLOSE = re.compile(r"^(.*)'$")


class Macro:
    name: str
    value: str

    def __init__(self, name: str, value: str = '') -> None:
        self.name = name
        self.value = value


defs: List[Macro] = []


def init() -> None:
    '''Reset definitions to defaults.'''
    # Initialize predefined macros.
    defs.clear()
    defs.append(Macro('--'))
    defs.append(Macro('--header-ids'))


def getValue(name: str) -> Optional[str]:
    '''Return named macro value or null if it doesn't exist.'''
    for mdef in defs:
        if mdef.name == name:
            return mdef.value
    return None


def setValue(name: str, value: str) -> None:
    '''Set named macro value or add it if it doesn't exist.
       If the name ends with '?' then don't set the macro if it already exists.
       `quote` is a single character: ' if a literal value, ` if an expression value.'''
    if options.skipMacroDefs():
        return  # Skip if a safe mode is set.
    existential = False
    if name.endswith('?'):
        name = name[:-1]  # Strip trailing '?'.
        existential = True
    if name == '--' and value != '':
        options.errorCallback('the predefined blank \'--\' macro cannot be redefined')
        return
    for mdef in defs:
        if mdef.name == name:
            if not existential:
                mdef.value = value
            return
    defs.append(Macro(name, value))


def render(text: str, silent: bool = False) -> str:
    '''Render macro invocations in text string.
        Render Simple invocations first, followed by Parametized, Inclusion and Exclusion invocations.'''
    MATCH_COMPLEX = re.compile(r'\\?{([\w\-]+)([!=|?](?:|.*?[^\\]))}',
                               re.DOTALL)  # Parametrized, Inclusion and Exclusion invocations.
    MATCH_SIMPLE = re.compile(r'\\?{([\w\-]+)()}')  # Simple macro invocation.
    result = text
    for find in [MATCH_SIMPLE, MATCH_COMPLEX]:
        def repl(match):
            if match[0].startswith('\\'):
                return match[0][1:]
            params = match[2]
            if params.startswith('?'):
                # DEPRECATED: Existential macro invocation.
                if not silent:
                    options.errorCallback(f'existential macro invocations are deprecated: {match[0]}')
                return match[0]
            name = match[1]
            value = getValue(name)  # Macro value is null if macro is undefined.
            if value is None:
                if not silent:
                    options.errorCallback(f'undefined macro: {match[0]}: {text}')
                return match[0]
            if find is MATCH_SIMPLE:
                return value
            params = params.replace(r'\}', '}')  # Unescape escaped} characters.
            if params[0] in '|':
                paramsList = params[1:].split('|')
                # Substitute macro parameters.
                # Matches macro definition formal parameters[$]$< param-number > [[\]: < default-param-value >$]
                # 1st group: [$]$
                # 2nd group: < param-number > (1, 2..)
                # 3rd group: [\]: < default-param-value >$
                # 4th group: < default-param-value >
                PARAM_RE = re.compile(r'\\?(\$\$?)(\d+)(\\?:(|.*?[^\\])\$)?', re.DOTALL)

                def repl(mr):
                    if mr[0].startswith('\\'):
                        # Unescape escaped macro parameters.
                        return mr[0][1:]
                    p1 = mr[1]
                    p2 = int(mr[2])
                    p3 = mr[3] or ''
                    p4 = mr[4] or ''
                    if p2 == 0:
                        return mr[0]  # $0 is not a valid parameter name.
                    # Unassigned parameters are replaced with a blank string.
                    param = '' if len(paramsList) < p2 else paramsList[p2 - 1]
                    if p3 != '':
                        if p3.startswith('\\'):
                            # Unescape escaped default parameter.
                            param += p3[1:]
                        elif param == '':
                            # Assign default parameter value.
                            param = p4
                            # Unescape escaped $ characters in the default value.
                            param = param.replace(r'\$', r'$')
                    if p1 == r'$$':
                        param = spans.render(param)
                    return param
                value = PARAM_RE.sub(repl, value)
                return value
            elif params[0] in '!=':  # Exclusion and inclusion macros.
                pattern = params[1:]
                skip: bool
                try:
                    skip = re.match(f'^{pattern}$', value) is None
                except:
                    if not silent:
                        options.errorCallback(f'illegal macro regular expression: {pattern}: {text}')
                    return match[0]
                if params[0] == '!':
                    skip = not skip
                return '\u0002' if skip else ''  # Flag line for deletion.
            else:
                options.errorCallback(f'illegal macro syntax: {match[0]}')
                return ''
        result = find.sub(repl, result)
    if '\u0002' in result:
        # Delete lines flagged by Inclusion/Exclusion macros.
        result = '\n'.join(filter(lambda line: '\u0002' not in line, result.split('\n')))
    return result



import re
from typing import Callable, List, Match, Optional, Pattern

from rimu import api, blockattributes, io, macros, options, utils
from rimu.expansion import Expand

MATCH_INLINE_TAG: Pattern[str] = re.compile(
    r'^(a|abbr|acronym|address|b|bdi|bdo|big|blockquote|br|cite|code|del|dfn|em|i|img|ins|kbd|mark|q|s|samp|small|span|strike|strong|sub|sup|time|tt|u|var|wbr)$',
    re.IGNORECASE)

# Custom type defintions.
Verify = Optional[Callable[[Match[str]], bool]]  # Additional match verification checks.
DelimiterFilter = Optional[Callable[[Match[str], 'Def'], str]]  # Return filtered delimiter content.
ContentFilter = Optional[Callable[[str, Match[str], Expand], str]]


class Def:
    '''Multi-line block element definition.'''
    openTag: str
    closeTag: str
    openMatch: Pattern[str]
    closeMatch: Pattern[str]  # $1 (if defined) is appended to block content.
    verify: Verify  # Additional match verification checks.
    delimiterFilter: DelimiterFilter  # Process opening delimiter. Return any delimiter content.
    contentFilter: ContentFilter
    expand: Expand
    name: str  # Optional unique identifier.

    def __init__(self,
                 openTag: str,
                 closeTag: str,
                 openMatch: Pattern[str],
                 closeMatch: Pattern[str] = None,
                 verify: Verify = None,
                 delimiterFilter: DelimiterFilter = None,
                 contentFilter: ContentFilter = None,
                 expand: Expand = None,
                 name: str = '',
                 ) -> None:
        self.openTag = openTag
        self.closeTag = closeTag
        self.openMatch = openMatch
        if closeMatch is not None:
            self.closeMatch = closeMatch
        else:
            self.closeMatch = self.openMatch
        self.verify = verify
        self.delimiterFilter = delimiterFilter
        self.contentFilter = contentFilter
        if expand is not None:
            self.expand = Expand.copyFrom(expand)
        else:
            self.expand = Expand()
        self.name = name

    @classmethod
    def copyFrom(cls, d: 'Def') -> 'Def':
        return Def(
            d.openTag,
            d.closeTag,
            d.openMatch,
            d.closeMatch,
            d.verify,
            d.delimiterFilter,
            d.contentFilter,
            d.expand,
            d.name,
        )


# Filter and Verify functions.

def htmlVerify(match: Match[str]) -> bool:
    '''Return False if the HTML tag is an inline (non-block) HTML tag.'''
    if match[2]:
        # Matched alphanumeric tag name.
        return MATCH_INLINE_TAG.search(match[2]) is None
    else:
         # Matched HTML comment or doctype tag.
        return True


def openingDelimiterFilter(match: Match[str], _) -> str:
    '''delimiterFilter that returns opening delimiter line text from match group $1.'''
    return match[1]


def classInjectionFilter(match: Match[str], d: Def) -> str:
    '''delimiterFilter for code, division and quote blocks.
       Inject $2 into block class attribute, set close delimiter to $1.'''
    p1 = match[2].strip()
    if p1:
        blockattributes.classes = p1
    d.closeMatch = re.compile('^' + re.escape(match[1]) + r'$')
    return ''


def macroDefContentFilter(text: str, match: Match[str], expand: Expand) -> str:
    '''contentFilter for multi-line macro definitions.'''
    m = re.search(r'^{([\w\-]+\??)}', match[0])  # Extract macro name from opening delimiter.
    assert m is not None
    name = m[1]
    text = re.sub(r"' *\\\n", "'\n", text)  # Unescape line-continuations.
    text = re.sub(r"(' *[\\]+)\\\n", lambda match: f'{match[1]}\n', text)  # Unescape escaped line-continuations.
    text = utils.replaceInline(text, expand)  # Expand macro invocations.
    macros.setValue(name, text)
    return ''


def indentedContentFilter(text: str, *_) -> str:
    '''Strip indent from start of each line.'''
    m = re.search(r'\S', text)
    assert m is not None
    first_indent = m.start()
    result = []
    for line in text.split('\n'):
        # Strip first line indent width or up to first non-space character.
        m = re.search(r'\S|$', line)
        assert m is not None
        indent = m.start()
        if indent > first_indent:
            indent = first_indent
        result.append(line[indent:])
    return '\n'.join(result)


def quoteParagraphContentFilter(text: str, *_) -> str:
    '''Strip leading > from start of each line and unescape escaped leading >.'''
    result = []
    for line in text.split('\n'):
        line = re.sub(r'^>', '', line)
        line = re.sub(r'^\\>', '>', line)
        result.append(line)
    return '\n'.join(result)


defs: List[Def] = []  # Mutable definitions initialized by DEFAULT_DEFS.

DEFAULT_DEFS: List[Def] = [
    # Delimited blocks cannot be escaped with a backslash.

    # Multi-line macro literal value definition.
    Def(
        openMatch=macros.DEF_OPEN,  # $1 is first line of macro.
        closeMatch=macros.DEF_CLOSE,
        openTag='',
        closeTag='',
        expand=Expand(macros=True),
        delimiterFilter=openingDelimiterFilter,
        contentFilter=macroDefContentFilter,
    ),
    # Comment block.
    Def(
        name='comment',
        openMatch=re.compile(r'^\\?\/\*+$'),
        closeMatch=re.compile(r'^\*+\/$'),
        openTag='',
        closeTag='',
        expand=Expand(
            skip=True, specials=True  # Fall-back if skip is disabled.
        ),
    ),
    # Division block.
    Def(
        name='division',
        openMatch=re.compile(
            r'^\\?(\.{2,})([\w\s-]*)$'),  # $1 is delimiter text, $2 is optional class names.
        openTag='<div>',
        closeTag='</div>',
        expand=Expand(
            container=True, specials=True  # Fall-back if container is disabled.
        ),
        delimiterFilter=classInjectionFilter,
    ),
    # Quote block.
    Def(
        name='quote',
        openMatch=re.compile(
            r'^\\?("{2,})([\w\s-]*)$'),  # $1 is delimiter text, $2 is optional class names.
        openTag='<blockquote>',
        closeTag='</blockquote>',
        expand=Expand(
            container=True, specials=True  # Fall-back if container is disabled.
        ),
        delimiterFilter=classInjectionFilter,
    ),
    # Code block.
    Def(
        name='code',
        openMatch=re.compile(
            r'^\\?(-{2,}|`{2,})([\w\s-]*)$'),  # $1 is delimiter text, $2 is optional class names.
        openTag='<pre><code>',
        closeTag='</code></pre>',
        expand=Expand(macros=False, specials=True),
        verify=lambda match:
            # The deprecated '-' delimiter does not support appended class names.
            not (match[1][0] == '-' and match[2].strip() != ''),
        delimiterFilter=classInjectionFilter,
    ),
    # HTML block.
    Def(
        name='html',
        # Block starts with HTML comment, DOCTYPE directive or block-level HTML start or end tag.
        # $1 is first line of block.
        # $2 is the alphanumeric tag name.
        openMatch=re.compile(
            r'^(<!--.*|<!DOCTYPE(?:\s.*)?|<\/?([a-z][a-z0-9]*)(?:[\s>].*)?)$',
            re.IGNORECASE),
        closeMatch=re.compile(r'^$'),  # Blank line or EOF.
        openTag='',
        closeTag='',
        expand=Expand(macros=True),
        verify=htmlVerify,
        delimiterFilter=openingDelimiterFilter,
        contentFilter=lambda text, *_: options.htmlSafeModeFilter(text),
    ),
    # Indented paragraph.
    Def(
        name='indented',
        openMatch=re.compile(r'^\\?(\s+\S.*)$'),  # $1 is first line of block.
        closeMatch=re.compile(r'^$'),  # Blank line or EOF.
        openTag='<pre><code>',
        closeTag='</code></pre>',
        expand=Expand(macros=False, specials=True),
        delimiterFilter=openingDelimiterFilter,
        contentFilter=indentedContentFilter,
    ),
    # Quote paragraph.
    Def(
        name='quote-paragraph',
        openMatch=re.compile(r'^\\?(>.*)$'),  # $1 is first line of block.
        closeMatch=re.compile(r'^$'),  # Blank line or EOF.
        openTag='<blockquote><p>',
        closeTag='</p></blockquote>',
        expand=Expand(
            macros=True,
            spans=True,
            specials=True  # Fall-back if spans is disabled.
        ),
        delimiterFilter=openingDelimiterFilter,
        contentFilter=quoteParagraphContentFilter,
    ),
    # Paragraph (lowest priority, cannot be escaped).
    Def(
        name='paragraph',
        openMatch=re.compile(r'(.*)'),  # $1 is first line of block.
        closeMatch=re.compile(r'^$'),  # Blank line or EOF.
        openTag='<p>',
        closeTag='</p>',
        expand=Expand(
            macros=True,
            spans=True,
            specials=True  # Fall-back if spans is disabled.
        ),
        delimiterFilter=openingDelimiterFilter),
]

# Reset definitions to defaults.


def init() -> None:
    '''Reset definitions to defaults.'''
    global defs
    # Make shallow of DEFAULT_DEFS (list and list objects).
    defs.clear()
    for d in DEFAULT_DEFS:
        defs.append(Def.copyFrom(d))


# If the next element in the reader is a valid delimited block render it
# and return True, else return False.
def render(reader: io.Reader, writer: io.Writer, allowed: List[str] = None) -> bool:
    global defs
    if allowed is None:
        allowed = []
    if reader.eof():
        options.panic('premature eof')
    for d in defs:
        if allowed and d.name not in allowed:
            continue
        match = d.openMatch.search(reader.cursor)
        if not match:
            continue
        # Escape non-paragraphs.
        if match[0][0] == '\\' and d.name != 'paragraph':
            # Drop backslash escape and continue.
            reader.cursor = reader.cursor[1:]
            continue
        if d.verify and not d.verify(match):
            continue
        # Process opening delimiter.
        delimiterText = d.delimiterFilter(match, d) if d.delimiterFilter else ''
        # Read block content into lines.
        lines: List[str] = []
        if delimiterText:
            lines.append(delimiterText)
        # Read content up to the closing delimiter.
        reader.next()
        content = reader.readTo(d.closeMatch or d.openMatch)
        if content is None:
            options.errorCallback('unterminated delimited block: ' + match[0])
        else:
            lines.extend(content)
        # Calculate block expansion options.
        expand = Expand.copyFrom(d.expand)
        expand.merge(blockattributes.opts)
        # Translate block.
        if expand.skip is not True:
            text = '\n'.join(lines)
            if d.contentFilter:
                text = d.contentFilter(text, match, expand)
            opentag = d.openTag
            if d.name == 'html':
                text = blockattributes.injectHtmlAttributes(text)
            else:
                opentag = blockattributes.injectHtmlAttributes(opentag)
            if expand.container:
                blockattributes.opts.container = None  # Consume before recursion.
                text = api.render(text)
            else:
                text = utils.replaceInline(text, expand)
            closetag = d.closeTag
            if d.name == 'division' and opentag == '<div>':
                # Drop div tags if the opening div has no attributes.
                opentag = ''
                closetag = ''
            writer.write(opentag)
            writer.write(text)
            writer.write(closetag)
            if not reader.eof() and (opentag + text + closetag):
                # Add a trailing '\n' if we've written a non-blank line and there are more source lines left.
                writer.write('\n')
        # Reset consumed Block Attributes expansion options.
        blockattributes.opts = Expand()
        return True
    return False  # No matching delimited block found.


def getDefinition(name: str) -> Optional[Def]:
    '''Return block definition or null if not found.'''
    for d in defs:
        if d.name == name:
            return d
    return None


def setDefinition(name: str, value: str) -> None:
    '''Update existing named definition.
       Value syntax: <open-tag>|<close-tag> block-options'''
    d = getDefinition(name)
    if not d:
        options.errorCallback(f"illegal delimited block name: {name}: |{name}|='{value}'")
        return
    match = re.search(r'^(?:(<[a-zA-Z].*>)\|(<[a-zA-Z/].*>))?(?:\s*)?([+-][ \w+-]+)?$', value.strip())
    if not match:
        options.errorCallback(f"illegal delimited block definition: |{name}|='{value}'")
        return
    if match[1] is not None:
        # Open and close tags are defined.
        d.openTag = match[1]
        d.closeTag = match[2]
    if match[3] is not None:
        d.expand.parse(match[3])



'''
 This module renders inline text containing Quote and Replacement elements.

 Quote and replacement processing involves splitting the source text into
 fragments where at the points where quotes and replacements occur then splicing fragments
 containing output markup into the breaks. A fragment is flagged as 'done' to
 exclude it from further processing.

 Once all quotes and replacements are processed fragments not yet flagged as
 'done' have special characters (&, <, >) replaced with corresponding special
 character entities. The fragments are then reassembled (defraged) into a
 resultant HTML string.
'''

import re
from typing import List, Match, Optional

from rimu import quotes, replacements, utils


class Fragment:
    text: str
    done: bool
    verbatim: str  # Replacements text rendered verbatim.

    def __init__(self,
                 text: str,
                 done: bool,
                 verbatim: str = '',
                 ):
        self.text = text
        self.done = done
        self.verbatim = verbatim


def render(source: str) -> str:
    result = preReplacements(source)
    fragments = [Fragment(text=result, done=False)]
    fragments = fragQuotes(fragments)
    fragSpecials(fragments)
    result = defrag(fragments)
    return postReplacements(result)


def defrag(fragments: List[Fragment]) -> str:
    '''Converts fragments to a string.'''
    result = ''
    for fragment in fragments:
        result += fragment.text
    return result


def fragQuotes(fragments: List[Fragment]) -> List[Fragment]:
    '''Fragment quotes in all fragments and return resulting fragments array.'''
    result: List[Fragment] = []
    for fragment in fragments:
        result.extend(fragQuote(fragment))
    # Strip backlash from escaped quotes in non-done fragments.
    for fragment in result:
        if not fragment.done:
            fragment.text = quotes.unescape(fragment.text)
    return result


def fragQuote(fragment: Fragment) -> List[Fragment]:
    '''Fragment quotes in a single fragment and return resulting fragments array.'''
    if fragment.done:
        return [fragment]
    # Find first matched quote in fragment text.
    quote: str
    match: Optional[Match[str]]
    startIndex: int = 0
    nextIndex: int = 0
    while True:
        match = quotes.quotesRe.search(fragment.text, nextIndex)
        if match is None:
            return [fragment]
        quote = match[1]
        # Check if quote is escaped.
        if match[0].startswith('\\'):
            # Restart search after escaped opening quote.
            nextIndex += match.start() + len(quote) + 1
            continue
        startIndex = match.start()
        nextIndex = match.end()
        break
    result: List[Fragment] = []
    # Arrive here if we have a matched quote.
    # The quote splits the input fragment into 5 or more output fragments:
    # Text before the quote, left quote tag, quoted text, right quote tag and text after the quote.
    qdef = quotes.getDefinition(match[1])
    assert qdef is not None
    # Check for same closing quote one character further to the right.
    quoted = match[2]
    while nextIndex < len(fragment.text) and fragment.text[nextIndex] == quote[0]:
        # Move to closing quote one character to right.
        quoted += quote[0]
        nextIndex += 1
    before = fragment.text[:startIndex]
    after = fragment.text[nextIndex:]
    result.append(Fragment(text=before, done=False))
    result.append(Fragment(text=qdef.openTag, done=True))
    if not qdef.spans:
        # Spans are disabled so render the quoted text verbatim.
        quoted = utils.replaceSpecialChars(quoted)
        quoted = quoted.replace('\u0000', '\u0001')  # Substitute verbatim replacement placeholder.
        result.append(Fragment(text=quoted, done=True))
    else:
        # Recursively process the quoted text.
        result.extend(fragQuote(Fragment(text=quoted, done=False)))
    result.append(Fragment(text=qdef.closeTag, done=True))
    # Recursively process the following text.
    result.extend(fragQuote(Fragment(text=after, done=False)))
    return result


# Stores placeholder replacement fragments saved by `preReplacements()` and restored by `postReplacements()`.
savedReplacements: List[Fragment] = []


def preReplacements(text: str) -> str:
    '''Return text with replacements replaced with a placeholder character (see `postReplacements()`):
       '\u0000' is placeholder for expanded replacement text.
       '\u0001' is placeholder for unexpanded replacement text (replacements that occur within quotes are rendered verbatim).'''
    savedReplacements.clear()
    fragments = fragReplacements([Fragment(text=text, done=False)])
    # Reassemble text with replacement placeholders.
    result = ''
    for fragment in fragments:
        if fragment.done:
            savedReplacements.append(fragment)  # Save replaced text.
            result += '\u0000'  # Placeholder for replaced text.
        else:
            result += fragment.text
    return result


def postReplacements(text: str) -> str:
    '''Replace replacements placeholders with replacements text from savedReplacements[].'''
    def func(match):
        fragment = savedReplacements.pop(0)
        return fragment.text if match[0] == '\u0000'else utils.replaceSpecialChars(fragment.verbatim)
    return re.sub(r'[\u0000\u0001]', func, text)


def fragReplacements(fragments: List[Fragment]) -> List[Fragment]:
    '''Fragment replacements in all fragments and return resulting fragments array.'''
    result = list(fragments)
    for rdef in replacements.defs:
        tmp: List[Fragment] = []
        for fragment in result:
            tmp.extend(fragReplacement(fragment, rdef))
        result = list(tmp)  # TODO: Is this necessary, why not: result=tmp?
    return result


def fragReplacement(fragment: Fragment,  rdef: replacements.Def) -> List[Fragment]:
    '''Fragment replacements in a single fragment for a single replacement definition.
       Return resulting fragments array.'''
    if fragment.done:
        return [fragment]
    match = rdef.match.search(fragment.text)
    if match is None:
        return [fragment]
    # Arrive here if we have a matched replacement.
    # The replacement splits the input fragment into 3 output fragments:
    # Text before the replacement, replaced text and text after the replacement.
    before = fragment.text[:match.start()]
    after = fragment.text[match.end():]
    result: List[Fragment] = []
    result.append(Fragment(text=before, done=False))
    replacement: str
    if match[0].startswith('\\'):
        # Remove leading backslash.
        replacement = utils.replaceSpecialChars(match[0][1:])
    else:
        if rdef.filter is None:
            replacement = utils.replaceMatch(match, rdef.replacement)
        else:
            replacement = rdef.filter(match, rdef)
    result.append(Fragment(text=replacement, done=True, verbatim=match[0]))
    # Recursively process the remaining text.
    result.extend(fragReplacement(Fragment(text=after, done=False), rdef))
    return result


def fragSpecials(fragments: List[Fragment]) -> None:
    '''Replace special characters in all non-done fragments.'''
    for fragment in filter(lambda fragment: not fragment.done, fragments):
        fragment.text = utils.replaceSpecialChars(fragment.text)



import re
from typing import List, Match, Optional, Pattern

from rimu import (blockattributes, delimitedblocks, expansion, io, lineblocks,
                  options, utils)
from rimu.expansion import Expand


class Def:
    match: Pattern[str]
    listOpenTag: str
    listCloseTag: str
    itemOpenTag: str
    itemCloseTag: str
    termOpenTag: str  # Definition lists only.
    termCloseTag: str  # Definition lists only.

    def __init__(self,
                 match: Pattern[str],
                 listOpenTag: str,
                 listCloseTag: str,
                 itemOpenTag: str,
                 itemCloseTag: str,
                 termOpenTag: str = '',
                 termCloseTag: str = '',
                 ):
        self.match = match
        self.listOpenTag = listOpenTag
        self.listCloseTag = listCloseTag
        self.itemOpenTag = itemOpenTag
        self.itemCloseTag = itemCloseTag
        self.termOpenTag = termOpenTag
        self.termCloseTag = termCloseTag


# Information about a matched list item element.
class ItemInfo:
    match: Match[str]
    listdef: 'Def'
    id: str  # List ID.


defs: List[Def] = [
    # Prefix match with backslash to allow escaping.

    # Unordered lists.
    # $1 is list ID $2 is item text.
    Def(
        match=re.compile(r'^\\?\s*(-|\+|\*{1,4})\s+(.*)$'),
        listOpenTag='<ul>',
        listCloseTag='</ul>',
        itemOpenTag='<li>',
        itemCloseTag='</li>'),
    # Ordered lists.
    # $1 is list ID $2 is item text.
    Def(
        match=re.compile(r'^\\?\s*(?:\d*)(\.{1,4})\s+(.*)$'),
        listOpenTag='<ol>',
        listCloseTag='</ol>',
        itemOpenTag='<li>',
        itemCloseTag='</li>'),
    # Definition lists.
    # $1 is term, $2 is list ID, $3 is definition.
    Def(
        match=re.compile(r'^\\?\s*(.*[^:])(:{2,4})(|\s+.*)$'),
        listOpenTag='<dl>',
        listCloseTag='</dl>',
        itemOpenTag='<dd>',
        itemCloseTag='</dd>',
        termOpenTag='<dt>',
        termCloseTag='</dt>'),
]

ids: List[str] = []  # Stack of open list IDs.


def render(reader: io.Reader, writer: io.Writer) -> bool:
    global ids
    if reader.eof():
        options.panic('premature eof')
    startItem: Optional[ItemInfo]
    startItem = matchItem(reader)
    if startItem is None:
        return False
    ids = []
    renderList(startItem, reader, writer)
    # ids should now be empty.
    if ids:
        options.panic('list stack failure')
    return True


def renderList(item: ItemInfo, reader: io.Reader, writer: io.Writer) -> Optional[ItemInfo]:
    global ids
    ids.append(item.id)
    writer.write(blockattributes.injectHtmlAttributes(item.listdef.listOpenTag))
    nextItem: Optional[ItemInfo]
    while True:
        nextItem = renderListItem(item, reader, writer)
        if nextItem is None or nextItem.id != item.id:
            # End of list or next item belongs to parent list.
            writer.write(item.listdef.listCloseTag)
            ids.pop()
            return nextItem
        item = nextItem


def renderListItem(item: ItemInfo, reader: io.Reader, writer: io.Writer) -> Optional[ItemInfo]:
    '''Render the current list item, return the next list item or null if there are no more items.'''
    global ids
    d = item.listdef
    match = item.match
    text: str
    if d.termOpenTag:  # => definition list.
        writer.write(blockattributes.injectHtmlAttributes(d.termOpenTag, consume=False))
        blockattributes.id=''
        text = utils.replaceInline(match[1], Expand(macros=True, spans=True))
        writer.write(text)
        writer.write(d.termCloseTag)
    writer.write(blockattributes.injectHtmlAttributes(d.itemOpenTag))
    # Process item text from first line.
    itemLines = io.Writer()
    text = match[match.re.groups]
    itemLines.write(text + '\n')
    # Process remainder of list item i.e. item text, optional attached block, optional child list.
    reader.next()
    attachedLines = io.Writer()
    blankLines: int
    attachedDone: bool = False
    nextItem: Optional[ItemInfo]
    while True:
        blankLines = consumeBlockAttributes(reader, attachedLines)
        if blankLines >= 2 or blankLines == -1:
            # EOF or two or more blank lines terminates list.
            nextItem = None
            break
        nextItem = matchItem(reader)
        if nextItem:
            if nextItem.id in ids:
                # Next item belongs to current list or a parent list.
                pass
            else:
                # Render child list.
                nextItem = renderList(nextItem, reader, attachedLines)
            break
        if attachedDone:
            break  # Multiple attached blocks are not permitted.
        if blankLines == 0:
            savedIds = ids
            ids = []
            if delimitedblocks.render(reader, attachedLines,
                                      allowed=['comment', 'code', 'division', 'html', 'quote']):
                attachedDone = True
            else:
                # Item body line.
                itemLines.write(reader.cursor + '\n')
                reader.next()
            ids = savedIds
        elif blankLines == 1:
            if delimitedblocks.render(reader, attachedLines, ['indented', 'quote-paragraph']):
                attachedDone = True
            else:
                break
    # Write item text.
    text = itemLines.toString().strip()
    text = utils.replaceInline(text, Expand(macros=True, spans=True))
    writer.write(text)
    # Write attachment and child list.
    writer.buffer.extend(attachedLines.buffer)
    # Close list item.
    writer.write(d.itemCloseTag)
    return nextItem

# Consume blank lines and Block Attributes.
# Return number of blank lines read or -1 if EOF.


def consumeBlockAttributes(reader: io.Reader, writer: io.Writer) -> int:
    blanks = 0
    while True:
        if reader.eof():
            return -1
        if lineblocks.render(reader, writer, allowed=['attributes']):
            continue
        if reader.cursor != '':
            return blanks
        blanks += 1
        reader.next()


# Check if the line at the reader cursor matches a list related element.
# Unescape escaped list items in reader.
# If it does not match a list related element return null.
def matchItem(reader: io.Reader) -> Optional[ItemInfo]:
    # Check if the line matches a List definition.
    if reader.eof():
        return None
    item = ItemInfo()  # ItemInfo factory.
    # Check if the line matches a list item.
    for d in defs:
        match = d.match.search(reader.cursor)
        if match is not None:
            if match[0][0] == '\\':
                reader.cursor = reader.cursor[1:]  # Drop backslash.
                return None
            item.match = match
            item.listdef = d
            # The second to last match group is the list ID.
            item.id = match[match.re.groups - 1]
            return item
    return None



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="rimu",
    version="11.3.0",
    author="Stuart Rackham",
    author_email="srackham@gmail.com",
    license="MIT",
    description="Rimu is a readable-text to HTML markup language inspired by AsciiDoc and Markdown.",
    keywords="rimu markdown asciidoc",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/srackham/rimu-py",
    project_urls={
        'Documentation': 'https://srackham.github.io/rimu/',
        'Source': 'https://github.com/srackham/rimu-py',
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    package_dir={'': 'src'},
    packages=setuptools.find_packages(where='src'),
    entry_points={
        'console_scripts': [
            'rimupy=rimuc.rimuc:main',
        ],
    },
)
